/*-
 * Copyright (c) 1999-2000 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: config.c,v 1.1.2.10 2001/01/02 11:55:28 simon Exp $";
#endif /* !lint */

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <regex.h>
#include <stdarg.h>

#include "config.h"

#include "common.h"
#include "db.h"
#include "debug.h"
#include "ipa.h"
#include "path.h"
#include "rules.h"



#define UPDATE_DB_DEF		(5 * MINUTE)	/* default value for the "update_db_time" parameter */

#define FILE_PERM_RUSR		(S_IRUSR)			/* 0400 */
#define FILE_PERM_RGRP		(S_IRGRP)			/* 0040 */
#define FILE_PERM_RUSR_RGRP	(FILE_PERM_RUSR|FILE_PERM_RGRP) /* 0440 */

#define DIR_PERM_RUSR		(S_IRUSR|S_IXUSR)		/* 0500 */
#define DIR_PERM_RGRP		(S_IRGRP|S_IXGRP)		/* 0050 */
#define DIR_PERM_RUSR_RGRP	(DIR_PERM_RUSR|DIR_PERM_RGRP)	/* 0550 */

#define FILE_PERM_DEF		FILE_PERM_RUSR	/* default file permissions */
#define DIR_PERM_DEF		DIR_PERM_RUSR	/* default directory permissions */


typedef enum {
	NONE_SECTION,
	RULE_SECTION,
	GLOBAL_SECTION,
	LIMIT_SECTION,
	EXPIRE_SECTION,
	STARTUP_SECTION,
	IF_LIMIT_SECTION,
	IF_NOLIMIT_SECTION,
	SHUTDOWN_SECTION,
	DEBUG_SECTION,
	REACH_SECTION
} SECTION;

char		*cfgfilename = CFGFILE;		/* default configuration file */

static u_int	update_db_time_global;		/* global { update_db_time } */
static u_int	append_db_time_global;		/* global { append_db_time } */
static u_quad_t	maxchunk_global;		/* global { maxchunk } */
static int	use_syslog = 1;			/* ==0 if -d switch was used */
static int	db_uid_is_set, db_gid_is_set, db_dir_is_set;
static int	db_uid_is_named = 0, db_gid_is_named;
static int	db_perm_is_set;
static PARSING_MODE	parsing_mode;

static struct ipac	ac;

static u_int	lineno;		/* number of line in the configuration file */
static SECTION	section, section_prev, section_top;
static int	global_section_is_set, debug_section_is_set;
static int	debug_ipfil_is_set = 0,	debug_limit_is_set = 0,
		debug_exec_is_set = 0, debug_time_is_set = 0, debug_lock_is_set = 0,
		lock_db_is_set = 0, lock_wait_time_is_set = 0;

static regex_t	reg_brace, reg_rule, reg_global, reg_ipfw,
		reg_update_db_time, reg_append_db_time, reg_emptyline, reg_time_val,
		reg_size_val, reg_ipfil, reg_ipfil_val, reg_acl, reg_acl_val,
		reg_limit, reg_expire, reg_startup, reg_exec, reg_byte_limit,
		reg_zero_time, reg_expire_time, reg_if_limit, reg_if_nolimit,
		reg_time_exp_val, reg_maxchunk, reg_shutdown, reg_info,
		reg_db_owner, reg_db_owner_val, reg_db_perm, reg_db_perm_val,
		reg_worktime, reg_worktime_val, reg_db_dir, reg_reach,
		reg_lock_db, reg_lock_wait_time, reg_yesno_val;

#ifdef __FreeBSD__
static int	debug_ipfw_is_set = 0;
static regex_t	reg_ipfw_val;
#endif

/* regex for ``debug'' section */
static regex_t	reg_debug, reg_debug_ipfw, reg_debug_ipfil, reg_debug_limit,
		reg_debug_exec, reg_debug_time, reg_debug_lock;

static int	build_config_regexes(void);
#ifdef __GNUC__
static void	Syslog(int, const char *, ...) __attribute__ ((format (printf, 2, 3)));
static void	line_err(const char *, const char *, ...) __attribute__ ((format (printf, 2, 3)));
#else
static void	Syslog(int, const char *, ...);
static void	line_err(const char *, const char *, ...);
#endif /* __GNUC__ */

static char	*size_str = NULL;


/*
 * Wrapper for syslog() function.
 * If use_syslog == 1 then syslog() is used,
 *                    else stderr stream is used.
 */
void
Syslog(int priority, const char *message, ...)
{
	int		errno_save = errno;
	va_list		ap;

	va_start(ap, message);
	if (use_syslog)
		vsyslog(priority, message, ap);
	else {
		fflush(stdout);
		if (strstr(message, "%m") == NULL)
			vfprintf(stderr, message, ap);
		else {
			vfprintf(stderr, message, ap);
			fprintf(stderr, "\b%s", strerror(errno_save));
		}
		fprintf(stderr, "\n");
	}
	va_end(ap);
}

/*
 * Output error message for some line in configuration file.
 */
static void
line_err(const char *line, const char *message, ...)
{
	va_list		ap;

	va_start(ap, message);
	if (line != NULL)
		Syslog(LOG_ERR, "%s line %u \"%s\"", cfgfilename, lineno, line);
	else
		Syslog(LOG_ERR, "%s line %u", cfgfilename, lineno);
	if (use_syslog)
		vsyslog(LOG_ERR, message, ap);
	else {
		fflush(stdout);
		vfprintf(stderr, message, ap);
		fprintf(stderr, "\n");
	}
	va_end(ap);
}

/*
 * Output error message about unknown format of some parameter.
 */
static void
wrong_format_msg(const char *line, const char *token, const char *message)
{
	char	*ch = *message == '\0' ? "" : ": ";

	switch (section) {
	case RULE_SECTION:
		line_err(line, "rule %s: wrong format for \"%s\"%s%s",
		    ac.rulename, token, ch, message);
		break;
	case LIMIT_SECTION:
		line_err(line, "rule %s, limit %s: wrong format for \"%s\"%s%s",
		    ac.rulename, ac.limit[ac.nlimit - 1].limitname, token, ch, message);
		break;
	case GLOBAL_SECTION:
		line_err(line, "global section: wrong format for \"%s\"%s%s",
		    token, ch, message);
		break;
	default:
		line_err(line, "wrong format for \"%s\"%s%s", token, ch, message);
	}
}

/*
 * Convert string to unsigned integer.
 * Wrapper for strtoul() function.
 *
 * XXX Asumed that sizeof(u_long) >= sizeof(u_int).
 */
static int
Strtoui(u_int *result, const char *nptr, char **endptr)
{
	u_long		val;
	int		errno_save = errno;

	errno = 0;
	val = strtoul(nptr, endptr, 10);
	if (val > UINT_MAX)
		errno = ERANGE;
	if (errno != 0) {
		line_err(nptr, "strtoul: %s", strerror(errno));
		return -1;
	}
	errno = errno_save;
	*result = val;
	return 0;
}

/*
 * Convert string to unsigned long long integer (unsigned quad).
 * Wrapper for strtouq() function.
 */
static int
Strtouq(u_quad_t *result, const char *nptr, char **endptr)
{
	u_quad_t	val;
	int		errno_save = errno;

	errno = 0;
	val = strtouq(nptr, endptr, 10);
	if (errno != 0) {
		line_err(nptr, "strtouq: %s", strerror(errno));
		return -1;
	}
	errno = errno_save;
	*result = val;
	return 0;
}

/*
 * Convert full string to unsigned long integer.
 */
static int
fullstrtoul(u_long *result, const char *nptr)
{
	char		*endptr;
	u_long		val;
	int		errno_save = errno;
	
	errno = 0;
	val = strtoul(nptr, &endptr, 10);
	if (errno != 0) {
		line_err(nptr, "strtoul: %s", strerror(errno));
		return -1;
	}
	errno = errno_save;
	if (*endptr != '\0')
		return -2;
	*result = val;
	return 0;
}

/*
 * Parse "rule" section: get rule name.
 */
static int
parse_rule(char *s)
{
	char		*s2;
	u_int		i;
	size_t		len;
	struct ipac	*acp;

	s = strchr(s, 'e') + 1;	/* skip "*rule" */
	s = skip_spaces(s);
	for (s2 = s; ; ++s2)
		if (*s2 == ' ' || *s2 == '\t' || *s2 == '{')
			break;
	len = s2 - s + 1;
	if ( (ac.rulename = calloc(len, sizeof(char))) == NULL) {
		Syslog(LOG_ERR, "calloc: %m");
		return -1;
	}
	memcpy(ac.rulename, s, len - 1);
	for (acp = ipac, i = 0; i < nipac; ++acp, ++i)
		if (strcmp(ac.rulename, acp->rulename) == 0) {
			line_err(s, "duplicated \"rule %s\" section", ac.rulename);
			return -1;
		}
	return 0;
}

#ifdef __FreeBSD__
/*
 * Parse "ipfw" parameter.
 */
static int
parse_ipfw(char *s)
{
	u_int		number, subnumber, i, found_number;
	int		action;
	char		*endptr;
	struct ipfwac	*ipfwacp;

	for (;;) {
		s = skip_spaces(s);
		if (!*s) {
			use_ipfw = 1;
			return 0;	/* EOL */
		}
		if (*s == '-') {
			action = SUB;
			++s;
		} else
			action = ADD;
		if (Strtoui(&number, s, &endptr) < 0)
			return -1;
		if (number > USHRT_MAX) {
			line_err(s, "rule %s: ipfw number should be less than %d",
			    ac.rulename, USHRT_MAX);
			return -1;
		}
		s = endptr;
		subnumber = 0;
		if (*s == '.')
			if (Strtoui(&subnumber, ++s, &endptr) < 0) {
				line_err((char *)0, "rule %s: ipfw subnumber should be less than %u",
				    ac.rulename, UINT_MAX);
				return -1;
			}
		s = endptr;
		for (i = found_number = 0, ipfwacp = ac.ipfwac; i < ac.nipfwac; ++ipfwacp, ++i)
			if (ipfwacp->number == number && ipfwacp->subnumber == subnumber) {
				found_number = 1;
				break;
			}
		if (found_number) {
			if (ac.ipfwac[i].action != action) {
				line_err((char *)0, "rule %s: ipfw rules %u.%u and -%u.%u have no effect",
				    ac.rulename, number, subnumber, number, subnumber);
				return -1;
			} else {
				line_err((char *)0, "rule %s: duplicated ipfw rule %u.%u (use only one)",
				     ac.rulename, number, subnumber);
				continue;
			}
		}
		if ( (ipfwacp = realloc(ac.ipfwac, (ac.nipfwac + 1) * sizeof *ipfwacp)) == NULL) {
			Syslog(LOG_ERR, "realloc: %m");
			return -1;
		}
		ac.ipfwac = ipfwacp;
		ipfwacp += ac.nipfwac;
		++ac.nipfwac;
		ipfwacp->number = (u_short)number;
		ipfwacp->subnumber = subnumber;
		ipfwacp->set = 0;
		ipfwacp->action = action;
	}
}
#endif /* __FreeBSD__ */

/*
 * Parse "ipfil" parameter.
 */
static int
parse_ipfil(char *s)
{
	int	action;
	u_int		i, j, found_number, found_list;
	u_int		number, group;
	char		type, *endptr;
	struct ipfilac		*ipfilacp;
	struct ipfilac_list	*list = NULL;

	for (;;) {
		s = skip_spaces(s);
		if (!*s) {
			use_ipfil = 1;
			return 0;	/* EOL */
		}
		if (*s == '-') {
			action = SUB;
			++s;
		} else
			action = ADD;
		type = *s;
		++s;
		if (*s == '@') {
			group = 0;
			++s;
		} else {
			if (Strtoui(&group, s, &endptr) < 0 || group > USHRT_MAX) {
				line_err((char *)0, "rule %s: ipfil group should be less than %u",
				    ac.rulename, USHRT_MAX);
				return -1;
			}
			s = endptr + 1;
		}
		if (Strtoui(&number, s, &endptr) < 0) {
			line_err((char *)0, "rule %s: ipfil rule number should be less than %u",
			    ac.rulename, UINT_MAX);
			return -1;
		}
		s = endptr;
		if (number == 0) {
			line_err((char *)0, "rule %s: ipfil number should be greater than zero", ac.rulename);
			return -1;
		}
		for (i = found_number = found_list = 0, ipfilacp = ac.ipfilac; i < ac.nipfilac; ++ipfilacp, ++i)
			if (ipfilacp->group == group && ipfilacp->type == type) {
				found_list = 1;
				for (list = ipfilacp->list, j = 0; j < ipfilacp->nlist; ++list, ++j)
					if (list->number == number) {
						found_number = 1;
						break;
					}
				break;
			}
		if (found_number) {
			if (list->action != action) {
				line_err((char *)0, "rule %s: ipfil rules %c%u@%u and -%c%u@%u have no effect",
				    ac.rulename, type, group, number, type, group, number);
				return -1;
			} else {
				line_err((char *)0, "rule %s: duplicated ipfil rule %c%u@%u (use only one)",
				    ac.rulename, type, group, number);
				continue;
			}
		}
		if (found_list) {
			/* list for group has been already created */
			if ( (list = realloc(ipfilacp->list, (ipfilacp->nlist + 1) * sizeof *list)) == NULL) {
				Syslog(LOG_ERR, "realloc: %m");
				return -1;
			}
			ipfilacp->list = list;
			list += ipfilacp->nlist;
			++ipfilacp->nlist;
		} else {
			/* create list for group */
			if ( (ipfilacp = realloc(ac.ipfilac, (ac.nipfilac + 1) * sizeof *ipfilacp)) == NULL) {
				Syslog(LOG_ERR, "realloc: %m");
				return -1;
			}
			ac.ipfilac = ipfilacp;
			ipfilacp += ac.nipfilac;
			++ac.nipfilac;
			if ( (ipfilacp->list = malloc(sizeof *ipfilacp->list)) == NULL) {
				Syslog(LOG_ERR, "malloc: %m");
				return -1;
			}
			ipfilacp->group = (u_short)group;
			ipfilacp->type = type;
			list = ipfilacp->list;
			ipfilacp->nlist = 1;
		}
		list->number = number;
		list->set = 0;
		list->action = action;
	}
}

/*
 * Parse parameter with time value.
 */
static int
parse_time(char *s, u_int *r)
{
	int	level = 0, err = 0;
	u_int	value, result = 0;
	char	*endptr;

	for (;;) {
		s = skip_spaces(s);
		if (!*s) {
			*r = result;
			return 0;	/* EOL */
		}
		if (Strtoui(&value, s, &endptr) < 0)
			return -1;
		s = endptr;
		switch (*s) {
		case 'h':
			if (level > 0)
				err = 1;
			else {
				level = 1;
				value *= HOUR;
			}
			break;
		case 'm':
			if (level > 1)
				err = 1;
			else {
				level = 2;
				value *= MINUTE;
			}
			break;
		default: /* 's' */
			if (level > 2)
				err = 1;
			else
				level = 3;
		}
		if (err) {
			line_err(s, "wrong time format");
			return -1;
		}
		if (result > UINT_MAX - value) {
			line_err(s, "too big value for time format");
			return -1;
		}
		result += value;
		++s;
	}
}

/*
 * Parse "update_db_time" parameter.
 */
static int
parse_update_db_time(char *s)
{
	u_int	result;

	if (parse_time(s, &result) < 0)
		return -1;
	if (result == 0) {
		line_err(s, "parameter \"update_db_time\" should be greater than zero");
		return -1;
	}
	if (section == RULE_SECTION)
		ac.update_db_time = result;
	else /* section == GLOBAL_SECTION */
		update_db_time_global = result;
	return 0;
}

/*
 * Parse "append_db_time" parameter.
 */
static int
parse_append_db_time(char *s)
{
	u_int	result;

	if (parse_time(s, &result) < 0)
		return -1;
	if (result == 0) {
		line_err(s, "parameter \"append_db_time\" should be greater than zero");
		return -1;
	}
	if (section == RULE_SECTION)
		ac.append_db_time = result;
	else /* section == GLOBAL_SECTION */
		append_db_time_global = result;
	return 0;
}

/*
 * Parse "size" parameter.
 *
 * XXX more tests should be done for (value * xBYTE).
 */
static int
parse_size(char *s)
{
	char		*p = s, *endptr;
	int		level = 0, err = 0;
	u_quad_t	value, result = 0;

	for (;;) {
		p = skip_spaces(p);
		if (!*p)
			break;		/* EOL */
		if (Strtouq(&value, p, &endptr) < 0)
			return -1;
		p = endptr;
		switch (*p) {
		case 't':
		case 'T':
			if (level > 0)
				err = 1;
			else {
				level = 1;
				value *= TBYTE;
			}
			break;
		case 'g':
		case 'G':
			if (level > 1)
				err= 1;
			else {
				level = 2;
				value *= GBYTE;
			}
			break;
		case 'm':
		case 'M':
			if (level > 2)
				err= 1;
			else {
				level = 3;
				value *= MBYTE;
			}
			break;
		case 'k':
		case 'K':
			if (level > 3)
				err= 1;
			else {
				level = 4;
				value *= KBYTE;
			}
			break;
		default: /* 'b' or 'B' */
			if (level > 4)
				err= 1;
			else
				level = 5;
		}
		if (err) {
			line_err(p, "wrong size format");
			return -1;
		}
		if (result > UQUAD_MAX - value) {
			line_err(p, "too big value for size format");
			return -1;
		}
		result += value;
		++p;
	}
	if (result == 0) {
		line_err(s, "parameter \"%s\" should be greater than zero",
		    section == LIMIT_SECTION ? "byte_limit" : "maxchunk");
		return -1;
	}
	switch (section) {
	case LIMIT_SECTION:
		ac.limit[ac.nlimit - 1].byte_limit = result;
		break;
	case RULE_SECTION:
		ac.maxchunk = result;
		break;
	default: /* section == NONE_SECTION */
		maxchunk_global = result;
	}
	return 0;
}

/*
 * Parse "limit" section: get limit name.
 */
static int
parse_limit(char *s)
{
	char		*s2;
	u_int		i;
	size_t		len;
	struct limit	*lp;

	s = strchr(s, 't') + 1;	/* skip "*limit" */
	s = skip_spaces(s);
	for (s2 = s; ; ++s2)
		if (*s2 == ' ' || *s2 == '\t' || *s2 == '{')
			break;
	len = s2 - s + 1;
	if ( (s2 = calloc(len, sizeof(char))) == NULL) {
		Syslog(LOG_ERR, "calloc: %m");
		return -1;
	}
	memcpy(s2, s, len - 1);
	for (i = 0, lp = ac.limit; i < ac.nlimit; ++lp, ++i)
		if (strcmp(s2, lp->limitname) == 0) {
			line_err(s, "rule %s: duplicated \"limit %s\" section",
			    ac.rulename, s2);
			free(s2);
			return -1;
		}
	if ( (lp = realloc(ac.limit, (ac.nlimit + 1) * sizeof *lp)) == NULL) {
		Syslog(LOG_ERR, "realloc: %m");
		return -1;
	}
	ac.limit = lp;
	lp = ac.limit + ac.nlimit;
	++ac.nlimit;
	lp->fp = NULL;
	lp->filename = NULL;
	lp->byte_limit = 0;
	lp->cmd = NULL;
	lp->ncmd = 0;
	lp->zero_time_param.upto = UPTO_NOTSET;
	lp->expire.time.upto = UPTO_NOTSET;
	lp->expire.cmd = NULL;
	lp->expire.ncmd = 0;
	lp->rc[0].cmd = lp->rc[0].cmd_if_limit = lp->rc[0].cmd_if_nolimit = NULL;
	lp->rc[0].ncmd = lp->rc[0].ncmd_if_limit = lp->rc[0].ncmd_if_nolimit = 0;
	lp->rc[1].cmd = lp->rc[1].cmd_if_limit = lp->rc[1].cmd_if_nolimit = NULL;
	lp->rc[1].ncmd = lp->rc[1].ncmd_if_limit = lp->rc[1].ncmd_if_nolimit = 0;
	lp->limitname = s2;
	lp->info = NULL;
	lp->pid = 0;
	lp->bcnt_sub = 0;
	lp->status = 0;
	lp->use_worktime = 0;
	for (i = 0; i < 7; ++i) {
		lp->worktime[i].interval = NULL;
		lp->worktime[i].ninterval = 0;
	}
	lp->is_changed = 0;
	return 0;
}

/*
 * Parse "expired time" value.
 */
static int
parse_time_exp(char *s)
{
	char		*p = s;
	char		*endptr;
	int		level = 0, err = 0;
	u_int		value;
	struct time_exp	*texp;
	struct limit	*lp = ac.limit + ac.nlimit - 1;

	texp = section == LIMIT_SECTION ? &lp->zero_time_param : &lp->expire.time;
	texp->seconds = 0;
	texp->monthes = 0;
	texp->upto = UPTO_SIMPLE;
	for (;;) {
		p = skip_spaces(p);
		if (*p == '+') {
			++p;
			switch (*p) {
			case UPTO_dAY:
			case UPTO_DAY:
				texp->upto = UPTO_DAY;
				break;
			case UPTO_wEEK:
			case UPTO_WEEK:
				texp->upto = UPTO_WEEK;
				break;
			default: /* all other UPTO_xxx have only one syntax */
				texp->upto = *p;
			}
			++p;
			texp->side = texp->seconds ? 1 : 0;
		}
		if (!*p)
			break;	/* EOL */
		if (Strtoui(&value, p, &endptr) < 0)
			return -1;
		p = endptr;
		switch (*p) {
		case 'M':
			if (level > 0)
				err = 1;
			else {
				level = 1;
				texp->monthes = value;
				value = 0;
			}
			break;
		case 'w':
		case 'W':
			if (level > 1)
				err = 1;
			else {
				level = 2;
				value *= WEEK;
			}
			break;
		case 'd':
		case 'D':
			if (level > 2)
				err = 1;
			else {
				level = 3;
				value *= DAY;
			}
			break;
		case 'h':
			if (level > 3)
				err = 1;
			else {
				level = 4;
				value *= HOUR;
			}
			break;
		case 'm':
			if (level > 4)
				err = 1;
			else {
				level = 5;
				value *= MINUTE;
			}
			break;
		default: /* 's' */
			if (level > 5)
				err = 1;
			else
				level = 6;
		}
		if (err) {
			line_err(p, "wrong time format");
			return -1;
		}
		if (texp->seconds > UINT_MAX - value) {
			line_err(p, "too big value for time format");
			return -1;
		}
		texp->seconds += value;
		++p;
	}
	if (section != EXPIRE_SECTION)
		if (texp->seconds == 0 && texp->monthes == 0 && texp->upto == UPTO_SIMPLE) {
			line_err(s, "parameter \"%s\" should be greater than zero",
			    section == LIMIT_SECTION ? "zero_time" : "expire_time");
			return -1;
		}
	return 0;
}

/*
 * Parse "exec" parameter.
 */
static int
parse_exec(char *q, char *s)
{
	char		*ptr, *start, bkp;
	short		uid_is_set = 0, uid_is_named = 0,
			gid_is_set = 0, gid_is_named = 0;
	u_int		*ncmdp_orig;
	u_long		uid, gid;
	struct limit	*lp;
	struct cmd	*cmdp, **cmdpp_orig;
	struct passwd	*pwd;
	struct group	*grp;
	
	for (ptr = q; *ptr != '='; ++ptr) {
		if (*ptr == '(') {
			start = ++ptr;
			if (*ptr != ':') {
				for (; *ptr != ')' && *ptr != ':'; ++ptr);
				bkp = *ptr;
				*ptr = '\0';
				if ( (pwd = getpwnam(start)) == NULL) {
					switch (fullstrtoul(&uid, start)) {
					case -1:
						return -1;
					case -2:
						line_err(q, "can't find user \"%s\"", start);
						return -1;
					default:
						break;
					}
				} else {
					uid = pwd->pw_uid;
					uid_is_named = 1;
				}
				if (pwd && bkp == ')') {
					gid = pwd->pw_gid;
					gid_is_set = 1;
					gid_is_named = 1;
				}
				*ptr = bkp;
				uid_is_set = 1;
			}
			if (*ptr == ':') {
				start = ++ptr;
				if (*start != ')') {
					for (; *ptr != ')'; ++ptr);
					*ptr = '\0';
					if ( (grp = getgrnam(start)) == NULL) {
						switch (fullstrtoul(&gid, start)) {
						case -1:
							return -1;
						case -2:
							line_err(q, "can't find group \"%s\"", start);
							return -1;
						default:
							break;
						}
					} else {
						gid = grp->gr_gid;
						gid_is_named = 1;
					}
					gid_is_set = 1;
				}
			}
			break;
		}
	}

	if (parsing_mode == RECONFIG_PARSING &&
	    (section == STARTUP_SECTION || section_prev == STARTUP_SECTION) )
		return 0;

	if (section == REACH_SECTION) {
		lp = ac.limit + ac.nlimit - 1;
		cmdpp_orig = &lp->cmd;
		ncmdp_orig = &lp->ncmd;
	} else if (section == EXPIRE_SECTION) {
		lp = ac.limit + ac.nlimit - 1;
		cmdpp_orig = &lp->expire.cmd;
		ncmdp_orig = &lp->expire.ncmd;
	} else if (section == STARTUP_SECTION) {
		if (section_top == RULE_SECTION) {
			cmdpp_orig = &ac.rc[0].cmd;
			ncmdp_orig = &ac.rc[0].ncmd;
		} else if (section_top == LIMIT_SECTION) {
			lp = ac.limit + ac.nlimit - 1;
			cmdpp_orig = &lp->rc[0].cmd;
			ncmdp_orig = &lp->rc[0].ncmd;
		} else /* section_top == NONE_SECTION */ {
			cmdpp_orig = &startup_global.cmd;
			ncmdp_orig = &startup_global.ncmd;
		}
	} else if (section == SHUTDOWN_SECTION) {
		if (section_top == RULE_SECTION) {
			cmdpp_orig = &ac.rc[1].cmd;
			ncmdp_orig = &ac.rc[1].ncmd;
		} else if (section_top == LIMIT_SECTION) {
			lp = ac.limit + ac.nlimit - 1;
			cmdpp_orig = &lp->rc[1].cmd;
			ncmdp_orig = &lp->rc[1].ncmd;
		} else /* section_top == NONE */ {
			cmdpp_orig = &shutdown_global.cmd;
			ncmdp_orig = &shutdown_global.ncmd;
		}
	} else if (section == IF_LIMIT_SECTION) {
		if (section_top == RULE_SECTION) {
			if (section_prev == STARTUP_SECTION) {
				cmdpp_orig = &ac.rc[0].cmd_if_limit;
				ncmdp_orig = &ac.rc[0].ncmd_if_limit;
			} else /* section_prev == SHUTDOWN_SECTION */ {
				cmdpp_orig = &ac.rc[1].cmd_if_limit;
				ncmdp_orig = &ac.rc[1].ncmd_if_limit;
			}
		} else /* section_top == LIMIT_SECTION */ {
			lp = ac.limit + ac.nlimit - 1;
			if (section_prev == STARTUP_SECTION) {
				cmdpp_orig = &lp->rc[0].cmd_if_limit;
				ncmdp_orig = &lp->rc[0].ncmd_if_limit;
			} else /* section_prev == SHUTDOWN_SECTION */ {
				cmdpp_orig = &lp->rc[1].cmd_if_limit;
				ncmdp_orig = &lp->rc[1].ncmd_if_limit;
			}
		}
	} else /* section == IF_NOLIMIT_SECTION */ {
		if (section_top == RULE_SECTION) {
			if (section_prev == STARTUP_SECTION) {
				cmdpp_orig = &ac.rc[0].cmd_if_nolimit;
				ncmdp_orig = &ac.rc[0].ncmd_if_nolimit;
			} else /* section_prev == SHUTDOWN_SECTION */ {
				cmdpp_orig = &ac.rc[1].cmd_if_nolimit;
				ncmdp_orig = &ac.rc[1].ncmd_if_nolimit;
			}
		} else /* section_top == LIMIT_SECTION */ {
			lp = ac.limit + ac.nlimit - 1;
			if (section_prev == STARTUP_SECTION) {
				cmdpp_orig = &lp->rc[0].cmd_if_nolimit;
				ncmdp_orig = &lp->rc[0].ncmd_if_nolimit;
			} else /* section_prev == SHUTDOWN_SECTION */ {
				cmdpp_orig = &lp->rc[1].cmd_if_nolimit;
				ncmdp_orig = &lp->rc[1].ncmd_if_nolimit;
			}
		}
	}
	if ( (cmdp = realloc(*cmdpp_orig, (*ncmdp_orig + 1) * sizeof *cmdp)) == NULL) {
		Syslog(LOG_ERR, "realloc: %m");
		return -1;
	}
	*cmdpp_orig = cmdp;
	cmdp = *cmdpp_orig + *ncmdp_orig;
	++*ncmdp_orig;
	cmdp->uid = uid;
	cmdp->gid = gid;
	cmdp->uid_is_set = uid_is_set; cmdp->uid_is_named = uid_is_named;
	cmdp->gid_is_set = gid_is_set; cmdp->gid_is_named = gid_is_named;
	s = skip_spaces(s);
	if ( (cmdp->str = calloc(strlen(s) + 1, sizeof(char))) == NULL) {
		Syslog(LOG_ERR, "calloc: %m");
		return -1;
	}
	memcpy(cmdp->str, s, strlen(s));
	return 0;
}

/*
 * Parse "acl" parameter.
 */
static int
parse_acl(char *s)
{
	char		*name_start, *acl_start, char_save;
	char		*aclp, **aclpp;
	u_long		val;
	size_t		len;

	if (section == GLOBAL_SECTION)
		aclpp = &acl_global;
	else /* section == RULE_SECTION */
		aclpp = &ac.acl;
	for (;;) {
		s = skip_spaces(s);
		if (!*s)
			break;	/* EOL */
		acl_start = s;
		if (*s == '!')
			++s;
		name_start = s;
		s = skip_chars(s);
		char_save = *s;
		*s = '\0';
		if (*name_start == '%') {
			++name_start;
			if (strcmp(name_start, "ALL%") != 0) {
				if (getgrnam(name_start) == NULL)
					switch (fullstrtoul(&val, name_start)) {
					case -1:
						return -1;
					case -2:
						line_err((char *)0, "can't find group \"%s\"", name_start);
						return -1;
					default:
						break;
					}
			}
		} else {
			if (getpwnam(name_start) == NULL)
				switch (fullstrtoul(&val, name_start)) {
				case -1:
					return -1;
				case -2:
					line_err((char *)0, "can't find user \"%s\"", name_start);
					return -1;
				default:
					break;
				}
		}
		len = *aclpp == NULL ? 0 : strlen(*aclpp);
		if ( (aclp = realloc(*aclpp, (len + s - acl_start + 2) * sizeof **aclpp)) == NULL) {
			Syslog(LOG_ERR, "realloc: %m");
			return -1;
		}
		*aclpp = aclp;
		aclp += len;
		*aclp = ' ';
		memcpy(aclp + 1, acl_start, s - acl_start + 1);
		len += s - acl_start + 1;
		*(aclp + len) = '\0';
		*s = char_save;
	}
	return 0;
}

/*
 * Parse "info" parameter.
 */
static int
parse_info(char *s)
{
	char	*info;

	if ( (info = calloc(strlen(s) + 1, sizeof(char))) == NULL) {
		Syslog(LOG_ERR, "calloc: %m");
		return -1;
	}
	strcpy(info, s);
	if (section == RULE_SECTION) {
		if (ac.info) {
			line_err(s, "rule %s: duplicated \"info\" parameter",
			    ac.rulename);
			free(info);
			return -1;
		}
		ac.info = info;
	} else /* section == LIMIT_SECTION */ {
		if (ac.limit[ac.nlimit - 1].info) {
			line_err(s, "rule %s limit %s: duplicated \"info\" parameter",
			    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
			free(info);
			return -1;
		}
		ac.limit[ac.nlimit - 1].info = info;
	}
	return 0;
}

/*
 * Parse "db_owner" parameter.
 */
static int
parse_db_owner(char *s)
{
	char		*start, bkp;
	char		*ptr = s;
	u_long		uid, gid;
	struct passwd	*pwd;
	struct group	*grp;

	db_uid_is_set = db_gid_is_set = 0;
	start = ptr;
	for (; *ptr && *ptr != ':'; ++ptr)
		;
	if (*start != ':') {
		bkp = *ptr;
		*ptr = '\0';
		if ( (pwd = getpwnam(start)) == NULL) {
			switch (fullstrtoul(&uid, start)) {
			case -1:
				return -1;
			case -2:
				line_err(s, "can't find user \"%s\"", start);
				return -1;
			default:
				break;
			}
		} else {
			db_uid_is_named = 1;
			uid = pwd->pw_uid;
		}
		if (pwd != NULL && bkp != ':') {
			gid = pwd->pw_gid;
			db_gid_is_set = 1;
			db_gid_is_named = 1;
		}
		*ptr = bkp;
		db_uid_is_set = 1;
	}
	if (*ptr == ':') {
		start = ++ptr;
		if (*start) {
			if ( (grp = getgrnam(start)) == NULL)
				switch (fullstrtoul(&gid, start)) {
				case -1:
					return -1;
				case -2:
					line_err(s, "can't find group \"%s\"", start);
					return -1;
				default:
					break;
				}
			else {
				db_gid_is_named = 1;
				gid = grp->gr_gid;
			}
			db_gid_is_set = 1;
		}
	}
	db_uid = uid;
	db_gid = gid;
	return 0;
}

/*
 * Parse "db_perm" parameter.
 */
static void
parse_db_perm(char *s)
{
	db_perm_is_set = 1;
	if ( (*s == 'u' && *(s + 1) == 'g') || (*s == 'g' && *(s + 1) == 'u') ) {
		db_fileperm = FILE_PERM_RUSR_RGRP;
		db_dirperm = DIR_PERM_RUSR_RGRP;
	} else if (*s == 'u') {
		db_fileperm = FILE_PERM_RUSR;
		db_dirperm = DIR_PERM_RUSR;
	} else /* *s == 'g' */ {
		db_fileperm = FILE_PERM_RGRP;
		db_dirperm = DIR_PERM_RGRP;
	}
}

/*
 * Parse "db_dir" parameter.
 */
static int
parse_db_dir(char *s)
{
	char	*ptr;

	db_dir_is_set = 0;
	if (*s != '/') {
		line_err(s, "directory in \"db_dir\" parameter should be given with absolute pathname");
		return -1;
	}
	if ( (ptr = calloc(strlen(s) + 1, sizeof(char))) == NULL) {
		Syslog(LOG_ERR, "calloc: %m");
		return -1;
	}
	strcpy(ptr, s);
	db_dir = ptr;
	db_dir_is_set = 1;
	return 0;
}

/*
 * Parse "worktime" parameter.
 */
static int
parse_worktime(char *s)
{
	int		wday, h1, m1, h2, m2, old_h2, old_m2;
	struct interval	*intp;
	struct worktime	*wtp;

	if (section == RULE_SECTION) {
		wtp = ac.worktime;
		ac.use_worktime = 1;
	} else if (section == LIMIT_SECTION) {
		wtp = ac.limit[ac.nlimit - 1].worktime;
		ac.limit[ac.nlimit - 1].use_worktime = 1;
	} else /* section == GLOBAL__SECTION */ {
		wtp = worktime_global;
		use_worktime_global = 1;
	}
	for (;;) {
		switch (*s) {
		case 's':
		case 'S':
			wday = 0;
			break;
		case 'm':
		case 'M':
			wday = 1;
			break;
		case 't':
		case 'T':
			wday = 2;
			break;
		case 'w':
		case 'W':
			wday = 3;
			break;
		case 'h':
		case 'H':
			wday = 4;
			break;
		case 'f':
		case 'F':
			wday = 5;
			break;
		default: /* 'a' || 'A' */
			wday = 6;
		}
		if (wtp[wday].interval != NULL) {
			wrong_format_msg(s, "worktime", "day should be specified only one time");
			return -1;
		}
		old_h2 = old_m2 = 0;
		++s;
		for (;;) {
			s = skip_spaces(s);
			if (!*s)
				return 0;	/* EOL */
			if (*s == '*') {
				h1 = m1 = 0;
				h2 = 24;
				m2 = 00;
			} else {
				if (isdigit(*s) == 0)
					break;
				if (sscanf(s, "%d:%d-%d:%d", &h1, &m1, &h2, &m2) != 4) {
					line_err(s, "sscanf(\"%%d:%%d-%%d:%%d\") failed: %m");
					return -1;
				}
				if (h1 > 23 || h2 > 23 || m1 > 59 || m2 > 59) {
					wrong_format_msg(s, "worktime", "wrong value of hours or minutes");
					return -1;
				}
				if ((h1 * HOUR + m1) > (h2 * HOUR + m2)) {
					wrong_format_msg(s, "worktime", "wrong time interval");
					return -1;
				}
				if ((old_h2 * HOUR + old_m2) > (h1 * HOUR + m1)) {
					wrong_format_msg(s, "worktime", "unsuccessive time interval");
					return -1;
				}
			}
			old_h2 = h2;
			old_m2 = m2;
			if ( (wtp[wday].interval = realloc(wtp[wday].interval,
			    wtp[wday].ninterval * sizeof(struct interval))) == NULL) {
				Syslog(LOG_ERR, "realloc: %m");
				return -1;
			}
			intp = wtp[wday].interval + wtp[wday].ninterval;
			intp->h1 = h1;
			intp->m1 = m1;
			intp->h2 = h2;
			intp->m2 = m2;
			++wtp[wday].ninterval;
			s = skip_chars(s);
		}
	}
}

/*
 * Parse "debug_*" parameter.
 */
int
parse_debug_level(int *result, char *s)
{
	int	level;

	if (sscanf(s, "%d", &level) != 1) {
		line_err(s, "sscanf(\"%%d\") failed: %m ");
		return -1;
	}
	*result = level;
	return 0;
}

/*
 * Parse "yes/no" parameter.
 */
static int
parse_yesno(char *s)
{
	return (*s == 'y' || *s == 'Y') ? 1 : 0;
}

static int
parse_lock_wait_time(char *s)
{
	if (parse_time(s, &lock_wait_time) < 0)
		return -1;
	if (lock_wait_time == 0) {
		wrong_format_msg(s, "lock_wait_time", "should be greater then 0 seconds");
		return -1;
	}
	lock_wait_time_is_set = 1;
	return 0;
}
/*
 * Skip variable and '=' sign from string.
 */
static char *
skip_var(register char *s)
{
	char	*ptr = strchr(s, '=');

	return ptr ? ptr + 1 : NULL;
}

#ifdef __FreeBSD__
/*
 * Compare two struct ipfwac.
 */
static int
cmp_ipfwac(const void *p1, const void *p2)
{
	const struct ipfwac	*q1 = p1, *q2 = p2;

	if (q1->number > q2->number)
		return 1;
	if (q1->number == q2->number) {
		if (q1->subnumber > q2->subnumber)
			return 1;
		return q1->subnumber < q2->subnumber ? -1 : 0;
	}
	return -1; 
}

/*
 * Sort ipfwac list for ac.
 */
static void
sort_ipfwac(void)
{
	qsort(ac.ipfwac, ac.nipfwac, sizeof *ac.ipfwac, cmp_ipfwac);
}
#endif /* __FreeBSD__ */

/*
 * Compare two struct ipfilac.
 */
static int
cmp_ipfilac(const void *p1, const void *p2)
{
	const struct ipfilac	*q1 = p1, *q2 = p2;

	if (q1->group > q2->group)
		return 1;
	if (q1->group == q2->group)
		return q1->type == 'o' ? -1 : 1;
	return -1;
}

/*
 * Compare two ipfilac lists.
 */
static int
cmp_ipfilac_list(const void *p1, const void *p2)
{
	const struct ipfilac_list	*q1 = p1, *q2 = p2;

	if (q1->number > q2->number)
		return 1;
	return q2->number < q2->number ? -1 : 0;
}

/*
 * Sort ipfilac lists for ac.
 */
static void
sort_ipfilac(void)
{
	u_int		i;
	struct ipfilac	*ipfilacp;

	qsort(ac.ipfilac, ac.nipfilac, sizeof *ac.ipfilac, cmp_ipfilac);
	for (i = 0, ipfilacp = ac.ipfilac; i < ac.nipfilac; ++ipfilacp, ++i)
		qsort(ipfilacp->list, ipfilacp->nlist, sizeof *ipfilacp->list, cmp_ipfilac_list);
}

/*
 * Set initial (default) values for ac.
 */
static void
reset_ac(void)
{
	u_int		i;

	ac.rulename = NULL;
	ac.update_db_time = 0;
	ac.append_db_time = 0;
	ac.fp = NULL;
	ac.bcnt_sub = 0;
#ifdef __FreeBSD__
	ac.ipfwac = NULL;
	ac.nipfwac = 0;
#endif
	ac.ipfilac = NULL;
	ac.nipfilac = 0;
	ac.limit = NULL;
	ac.nlimit = 0;
	ac.rc[0].cmd = ac.rc[0].cmd_if_limit = ac.rc[0].cmd_if_nolimit = NULL;
	ac.rc[0].ncmd = ac.rc[0].ncmd_if_limit = ac.rc[0].ncmd_if_nolimit = 0;
	ac.rc[1].cmd = ac.rc[1].cmd_if_limit = ac.rc[1].cmd_if_nolimit = NULL;
	ac.rc[1].ncmd = ac.rc[1].ncmd_if_limit = ac.rc[1].ncmd_if_nolimit = 0;
	ac.acl = NULL;
	ac.maxchunk = 0;
	ac.info = NULL;
	ac.wakeup = 0;
	ac.use_worktime = 0;
	for (i = 0; i < 7; ++i) {
		ac.worktime[i].interval = NULL;
		ac.worktime[i].ninterval = 0;
	}
}

/*
 * Output error message for unexpected parameter or section in
 * configuration file.
 */
static void
isnot_expected_msg(const char *msg)
{
	line_err((char *)0, "\"%s\" is not expected here", msg);
}


/*
 * Check security for configuration file: absolute path, regular file,
 * owned by UID 0 (root), writeable only for user (root),
 * show warning message if file is readable by group or other user.
 */
static int
check_cfgfilename(void)
{
	struct stat	statbuf;

	if (*cfgfilename != '/') {
		syslog(LOG_ERR, "configuration file should be given with absolute pathname");
		return -1;
	}
	if (stat(cfgfilename, &statbuf) < 0) {
		syslog(LOG_ERR, "stat(%s): %m", cfgfilename);
		return -1;
	}
	if (!S_ISREG(statbuf.st_mode)) {
		syslog(LOG_ERR, "%s should be regular file", cfgfilename);
		return -1;
	}
	if (statbuf.st_uid != 0) {
		syslog(LOG_ERR, "file %s should be owned by root", cfgfilename);
		return -1;
	}
	if (statbuf.st_mode & (S_IWGRP|S_IWOTH)) {
		syslog(LOG_ERR, "file %s should not have write permissions for group or other users",
		    cfgfilename);
		return -1;
	}
	if (statbuf.st_mode & (S_IROTH|S_IROTH))
		syslog(LOG_WARNING,  "file %s is readable by group and/or other users", cfgfilename);
	return 0;
}

/*
 * Read line from fp, return number of characters in line or
 * -1 if error occured. *size is equal to #bytes previously allocated in *strp
 * by malloc(). If there is not enough space in *strp, then readline()
 * reallocates *strp and updates *size.
 *
 * This is wrapper for fgets() function with some extensions.
 */
int
readline(char **strp, size_t *size, FILE *fp, char *filename)
{
#define CHUNK_SIZE 100

/*
 * XXX CHUNK_SIZE can't be less then 1, because following code doesn't
 *     work with such value.
 */

#if (CHUNK_SIZE <= 1)
# error Macro CHUNK_SIZE should be greater than 1
#endif

	char		*s;
	size_t		can_nread;	/* #bytes to read in next fgets() */
	size_t		len;		/* length of read string */
	size_t		nread = 0;	/* how many bytes we read */

	if (*strp == NULL) {
		if ( (*strp = malloc(CHUNK_SIZE)) == NULL) {
			Syslog(LOG_ERR, "malloc(%u bytes): %m", CHUNK_SIZE);
			return -1;
		}
		*size = CHUNK_SIZE;
	}
	for (s = *strp, can_nread = *size; ;) {
		if (fgets(s, can_nread, fp) == NULL) {
			if (feof(fp) != 0) {
				++lineno;
				break;	/* EOF */
			}
			Syslog(LOG_ERR, "fgets(%s): %m", filename);
			return -1;
		}
		len = strlen(s);
		if (len > 0) {
			nread += len;
			if (s[len - 1] == '\n') {
				++lineno;
				if (len >= 2) {
					if (s[len - 2] == '\\') {
						s += len - 2;
						nread -= 2;
						can_nread -= len - 2;
						continue;
					} else
						break;
				} else
					break;
			}
		}
		if ( (s = realloc(*strp, *size + CHUNK_SIZE)) == NULL) {
			Syslog(LOG_ERR, "realloc(%u bytes): %m", *size + CHUNK_SIZE);
			return -1;
		}
		*strp = s;
		*size += can_nread = CHUNK_SIZE;
		s += nread;
	}
	return nread;
}

/*
 * Main function for parsing whole configuration file.
 */
int
parse_config(int mode) 
{
	char		*buf = NULL, *string, *sp;
	short		startup_global_flag = 0, shutdown_global_flag = 0,
			startup_rule_flag = 0, shutdown_rule_flag = 0,
			startup_if_limit_rule_flag = 0, startup_if_nolimit_rule_flag = 0,
			shutdown_if_limit_rule_flag = 0, shutdown_if_nolimit_rule_flag = 0,
			startup_limit_flag = 0, shutdown_limit_flag = 0,
			startup_if_limit_limit_flag = 0, startup_if_nolimit_limit_flag = 0,
			shutdown_if_limit_limit_flag = 0, shutdown_if_nolimit_limit_flag = 0,
			expire_flag = 0, reach_flag = 0;
	size_t		bufsize;
	u_int		i;
	int		len;
	FILE		*fp;
	struct ipac	*ipacp;

	if (mode == TEST_PARSING)
		use_syslog = 0;
	else if (check_cfgfilename() < 0)
		return -1;

	reset_ac();
	update_db_time_global = 0;
	append_db_time_global = 0;
	maxchunk_global = 0;
	acl_global = NULL;
	db_uid_is_set = db_gid_is_set = db_dir_is_set = db_perm_is_set = 0;
	global_section_is_set = debug_section_is_set = 0;
	lock_db_is_set = lock_wait_time_is_set = 0;
	lock_db = 1;
	lock_wait_time = 0;
	reset_debug();
	startup_global.cmd = startup_global.cmd_if_limit = startup_global.cmd_if_nolimit = NULL;
	startup_global.ncmd = startup_global.ncmd_if_limit = startup_global.ncmd_if_nolimit = 0;
	shutdown_global.cmd = shutdown_global.cmd_if_limit = shutdown_global.cmd_if_nolimit = NULL;
	shutdown_global.ncmd = shutdown_global.ncmd_if_limit = shutdown_global.ncmd_if_nolimit = 0;

	for (i = 0; i < 7; ++i)
		worktime_global[i].interval = NULL;

	section = section_top = section_prev = NONE_SECTION;

	parsing_mode = mode;
	if (build_config_regexes() < 0) {
		Syslog(LOG_ERR, "can't build all regular expressions for configuration file parsing");
		return -1;
	}
	if ( (fp = fopen(cfgfilename, "r")) == NULL) {
		Syslog(LOG_ERR, "fopen(%s, \"r\"): %m", cfgfilename);
		return -1;
	}

	lineno = 0;

	for (;;) {
		len = readline(&buf, &bufsize, fp, cfgfilename);
		if (len == 0 && feof(fp) != 0) {
			if (section != NONE_SECTION) {
				line_err(buf, "unexpected end of configuration file");
				goto parsing_failed;
			}
			if (fclose(fp) != 0) {
				Syslog(LOG_ERR, "fclose(%s): %m", cfgfilename);
				return -1;
			}
			if (buf)
				free(buf);
			if (mode != TEST_PARSING) {
				for (i = 0, ipacp = ipac; i < nipac; ++ipacp, ++i) {
					if (ipacp->append_db_time == 0)
						ipacp->append_db_time = append_db_time_global;
					if (ipacp->maxchunk == 0)
						ipacp->maxchunk = maxchunk_global;
					if (ipacp->update_db_time > DAY ||
					    ipacp->append_db_time > DAY)
						syslog(LOG_INFO, "rule %s: \"update_db_time\" or \"append_db_time\" is greater than %u seconds (1 day), it doesn't make sense",
						    ipacp->rulename, DAY);
					if (ipacp->update_db_time == 0)
						ipacp->update_db_time = update_db_time_global ? update_db_time_global : UPDATE_DB_DEF;
					if (ipacp->append_db_time != 0 && ipacp->update_db_time > ipacp->append_db_time)
						syslog(LOG_INFO, "rule %s: \"append_db_time\" is less than \"update_db_time\", it doesn't make sense",
						    ipacp->rulename);
				}
			}
			if (!db_uid_is_set)
				db_uid = myuid;
			if (!db_gid_is_set)
				db_gid = mygid;
			if (!db_perm_is_set) {
				db_fileperm = FILE_PERM_DEF;
				db_dirperm = DIR_PERM_DEF;
			}
			return 0;
		}
		if (len < 0) {
			Syslog(LOG_ERR, "can't read from file %s", cfgfilename);
			goto parsing_failed;
		}
		if (len == 0)
			continue;
		if (buf[len - 1] == '\n')
			/* remove new line character '\n' */
			buf[len - 1] = '\0';
		if (buf[0] == '#' || buf[0] == ';')
			continue;	/* comment */
		if (REGEXEC(emptyline, buf) == 0)
			continue;	/* empty line */
		string = skip_spaces(buf);
		remove_trailing_spaces(string);
		if (section == NONE_SECTION) {
			if (REGEXEC(rule, string) == 0) {
				section = RULE_SECTION;
				reset_ac();
				startup_rule_flag = shutdown_rule_flag =
				    startup_if_limit_rule_flag =
				    startup_if_nolimit_rule_flag =
				    shutdown_if_limit_rule_flag =
				    shutdown_if_nolimit_rule_flag = 0;
				if (parse_rule(string) < 0)
					goto parsing_failed;
			} else if (REGEXEC(global, string) == 0) {
				if (global_section_is_set != 0) {
					line_err(string, "duplicated \"global\" section");
					goto parsing_failed;
				}
				global_section_is_set = 1;
				section = GLOBAL_SECTION;
			} else if (REGEXEC(debug, string) == 0) {
				if (debug_section_is_set != 0) {
					line_err(string, "duplicated \"debug\" section");
					goto parsing_failed;
				}
				debug_section_is_set = 1;
				section = DEBUG_SECTION;
			} else if (REGEXEC(startup, string) == 0) {
				if (startup_global_flag != 0) {
					line_err(string, "duplicated \"startup\" section");
					goto parsing_failed;
				}
				startup_global_flag = 1;
				section = STARTUP_SECTION;
			} else if (REGEXEC(shutdown, string) == 0) {
				if (shutdown_global_flag != 0) {
					line_err(string, "duplicated \"shutdown\" section");
					goto parsing_failed;
				}
				shutdown_global_flag = 1;
				section = SHUTDOWN_SECTION;
			} else {
				line_err(buf, "expected begin of \"rule\", \"global\", \"startup\" or \"shutdown\" section");
				goto parsing_failed;
			}
			continue;
		}
		if (REGEXEC(brace, string) == 0) {
			if (section == RULE_SECTION) {
#ifdef __FreeBSD__
				if (!ac.ipfwac && !ac.ipfilac) {
					Syslog(LOG_ERR, "%s", cfgfilename);
					Syslog(LOG_ERR, "rule %s: rule should has \"ipfw\" and/or \"ipfil\" parameter",
					    ac.rulename);
#else
				if (!ac.ipfilac) {
					Syslog(LOG_ERR, "%s", cfgfilename);
					Syslog(LOG_ERR, "rule %s: rule should has \"ipfil\" parameter",
					    ac.rulename);
#endif /* __FreeBSD__ */
					goto parsing_failed;
				}
				if ( (ipacp = ipac_new()) == NULL)
					goto parsing_failed;
#ifdef __FreeBSD__
				if (ac.ipfwac) 
					sort_ipfwac();
#endif
				if (ac.ipfilac)
					sort_ipfilac();
				memcpy(ipacp, &ac, sizeof *ipacp);
				section = NONE_SECTION;
			} else if (section == GLOBAL_SECTION) {
				if (mode != TEST_PARSING && lock_wait_time_is_set && lock_db == 0)
					Syslog(LOG_WARNING, "global section: \"lock_wait_time\" parameter is used and database locking feature is disabled");
				section = NONE_SECTION;
			} else if (section == LIMIT_SECTION) {
				if (ac.limit[ac.nlimit - 1].byte_limit == 0) {
					Syslog(LOG_ERR, "%s", cfgfilename);
					Syslog(LOG_ERR, "rule %s limit %s: parameter \"byte_limit\" in \"limit\" section should be present",
					    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
					goto parsing_failed;
				}
				section = RULE_SECTION;
			} else if (section == EXPIRE_SECTION) {
				if (ac.limit[ac.nlimit - 1].expire.time.upto == UPTO_NOTSET) {
					Syslog(LOG_ERR, "%s", cfgfilename);
					Syslog(LOG_ERR, "rule %s limit %s: parameter \"expire_time\" in \"expire\" section should be present",
					    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
					goto parsing_failed;
				}
				section = LIMIT_SECTION;
			} else if (section == STARTUP_SECTION) {
				section = section_top;
				section_top = NONE_SECTION;
			} else if (section == SHUTDOWN_SECTION) {
				section = section_top;
				section_top = NONE_SECTION;
			} else if (section == REACH_SECTION) {
				section = LIMIT_SECTION;
			} else /* section == IF_LIMIT_SECTION || section == IF_NOLIMIT_SECTION */ {
				section = section_prev;
				section_prev = NONE_SECTION;
			}
			continue;
		}

		if (REGEXEC(limit, string) == 0) {
			if (section != RULE_SECTION) {
				isnot_expected_msg("limit");
				goto parsing_failed;
			}
			startup_limit_flag = shutdown_limit_flag =
			    startup_if_limit_limit_flag =
			    startup_if_nolimit_limit_flag =
			    shutdown_if_limit_limit_flag =
			    shutdown_if_nolimit_limit_flag = expire_flag = 0;
			section = LIMIT_SECTION;
			if (parse_limit(string) < 0)
				goto parsing_failed;
			continue;
		}
		if (REGEXEC(expire, string) == 0) {
			if (section != LIMIT_SECTION) {
				isnot_expected_msg("expire");
				goto parsing_failed;
			}
			if (expire_flag) {
				line_err(string, "rule %s limit %s: duplicated \"expire\" section",
				    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
				goto parsing_failed;
			}
			expire_flag = 1;
			section = EXPIRE_SECTION;
			continue;
		}
		if (REGEXEC(reach, string) == 0) {
			if (section != LIMIT_SECTION) {
				isnot_expected_msg("reach");
				goto parsing_failed;
			}
			if (reach_flag) {
				line_err(string, "rule %s limit %s: duplicated \"reach\" section",
				    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
				goto parsing_failed;
			}
			reach_flag = 0;
			section = REACH_SECTION;
			continue;
		}
		if (REGEXEC(startup, string) == 0) {
			if (section != RULE_SECTION && section != LIMIT_SECTION) {
				isnot_expected_msg("startup");
				goto parsing_failed;
			}
			if (section == RULE_SECTION) {
				if (startup_rule_flag) {
					line_err(string, "rule %s: duplicated \"startup\" section",
					    ac.rulename);
					goto parsing_failed;
				}
				startup_rule_flag = 1;
			} else /* section == LIMIT_SECTION */ {
				if (startup_limit_flag) {
					line_err(string, "rule %s limit %s: duplicated \"startup\" section",
					    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
					goto parsing_failed;
				}
				startup_limit_flag = 1;
			}
			section_top = section;
			section = STARTUP_SECTION;
			continue;
		}
		if (REGEXEC(shutdown, string) == 0) {
			if (section != RULE_SECTION && section != LIMIT_SECTION) {
				isnot_expected_msg("shutdown");
				goto parsing_failed;
			}
			if (section == RULE_SECTION) {
				if (shutdown_rule_flag) {
					line_err(string, "rule %s: duplicated \"shutdown\" section",
					    ac.rulename);
					goto parsing_failed;
				}
				shutdown_rule_flag = 1;
			} else /* section == LIMIT_SECTION */ {
				if (shutdown_limit_flag) {
					line_err(string, "rule %s limit %s: duplicated \"shutdown\" section",
					    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
					goto parsing_failed;
				}
				shutdown_limit_flag = 1;
			}
			section_top = section;
			section = SHUTDOWN_SECTION;
			continue;
		}
		if (REGEXEC(if_limit, string) == 0) {
			if ((section != STARTUP_SECTION && section != SHUTDOWN_SECTION) ||
			    section_top == NONE_SECTION) {
				isnot_expected_msg("if_limit_is_reached");
				goto parsing_failed;
			}
			if (section_top == RULE_SECTION) {
				if (section == STARTUP_SECTION) {
					if (startup_if_limit_rule_flag) {
						line_err(string, "rule %s: duplicated \"if_limit_is_reached\" section in \"startup\" section",
						    ac.rulename);
						goto parsing_failed;
					}
					startup_if_limit_rule_flag = 1;
				}
				if (section == SHUTDOWN_SECTION) {
					if (shutdown_if_limit_rule_flag) {
						line_err(string, "rule %s: duplicated \"if_limit_is_reached\" section in \"shutdown\" section",
						    ac.rulename);
						goto parsing_failed;
					}
					shutdown_if_limit_rule_flag = 1;
				}
			} else /* section_top == LIMIT_SECTION */ {
				if (section == STARTUP_SECTION) {
					if (startup_if_limit_limit_flag) {
						line_err(string, "rule %s limit %s: duplicated \"if_limit_is_reached\" section in \"startup\" section",
						    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
						goto parsing_failed;
					}
					startup_if_limit_limit_flag = 1;
				}
				if (section == SHUTDOWN_SECTION) {
					if (shutdown_if_limit_limit_flag) {
						line_err(string, "rule %s limit %s: duplicated \"if_limit_is_reached\" section in \"shutdown\" section",
						    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
						goto parsing_failed;
					}
					shutdown_if_limit_limit_flag = 1;
				}
			}
			section_prev = section;
			section = IF_LIMIT_SECTION;
			continue;
		}
		if (REGEXEC(if_nolimit, string) == 0) {
			if ((section != STARTUP_SECTION && section != SHUTDOWN_SECTION) ||
			    section_top == NONE_SECTION) {
				isnot_expected_msg("if_limit_is_not_reached");
				goto parsing_failed;
			}
			if (section_top == RULE_SECTION) {
				if (section == STARTUP_SECTION) {
					if (startup_if_nolimit_rule_flag) {
						line_err(string, "rule %s: duplicated \"if_limit_is_not_reached\" section in \"startup\" section",
						    ac.rulename);
						goto parsing_failed;
					}
					startup_if_nolimit_rule_flag = 1;
				}
				if (section == SHUTDOWN_SECTION) {
					if (shutdown_if_nolimit_rule_flag) {
						line_err(string, "rule %s: duplicated \"if_limit_is_not_reached\" section in \"shutdown\" section",
						    ac.rulename);
						goto parsing_failed;
					}
					shutdown_if_nolimit_rule_flag = 1;
				}
			} else /* section_top == LIMIT_SECTION */ {
				if (section == STARTUP_SECTION) {
					if (startup_if_nolimit_limit_flag) {
						line_err(string, "rule %s limit %s: duplicated \"if_limit_is_not_reached\" section in \"startup\" section",
						    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
						goto parsing_failed;
					}
					startup_if_nolimit_limit_flag = 1;
				}
				if (section == SHUTDOWN_SECTION) {
					if (shutdown_if_nolimit_limit_flag) {
						line_err(string, "rule %s limit %s: duplicated \"if_limit_is_not_reached\" section in \"shutdown\" section",
						    ac.rulename, ac.limit[ac.nlimit - 1].limitname);
						goto parsing_failed;
					}
					shutdown_if_nolimit_limit_flag = 1;
				}
			}
			section_prev = section;
			section = IF_NOLIMIT_SECTION;
			continue;
		}
		if ( (sp = skip_var(string)) == NULL) {
			line_err(string, "unknown syntax, expected parameter or beginning or ending of a section");
			goto parsing_failed;
		}
		sp = skip_spaces(sp);
		if (REGEXEC(emptyline, sp) == 0) {
			line_err(string, "expected value for parameter");
			goto parsing_failed;
		}
		if (REGEXEC(ipfw, string) == 0) {
			if (section != RULE_SECTION) {
				isnot_expected_msg("ipfw");
				goto parsing_failed;
			}
#ifdef __FreeBSD__
			if (REGEXEC(ipfw_val, sp) != 0) {
				wrong_format_msg(sp, "ipfw", "");
				goto parsing_failed;
			}
			if (parse_ipfw(sp) < 0)
				goto parsing_failed;
#else
			line_err(string, "IP Firewall is not supported. Don't use \"ipfw\" parameter.");
			goto parsing_failed;
#endif /* __FreeBSD__ */
		} else if (REGEXEC(ipfil, string) == 0) {
			if (section != RULE_SECTION) {
				isnot_expected_msg("ipfil");
				goto parsing_failed;
			}
			if (REGEXEC(ipfil_val, sp) != 0) {
				wrong_format_msg(sp, "ipfil", "");
				goto parsing_failed;
			}
			if (parse_ipfil(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(update_db_time, string) == 0) {
			if (section != RULE_SECTION && section != GLOBAL_SECTION) {
				isnot_expected_msg("update_db_time");
				goto parsing_failed;
			}
			if (REGEXEC(time_val, sp) != 0) {
				wrong_format_msg(sp, "update_db_time", "");
				goto parsing_failed;
			}
			if (parse_update_db_time(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(append_db_time, string) == 0) {
			if (section != RULE_SECTION && section != GLOBAL_SECTION) {
				isnot_expected_msg("append_db_time");
				goto parsing_failed;
			}
			if (REGEXEC(time_val, sp) != 0) {
				wrong_format_msg(sp, "append_db_time", "");
				goto parsing_failed;
			}
			if (parse_append_db_time(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(byte_limit, string) == 0) {
			if (section != LIMIT_SECTION) {
				isnot_expected_msg("byte_limit");
				goto parsing_failed;
			}
			if (REGEXEC(size_val, sp) != 0) {
				wrong_format_msg(sp, "byte_limit", "");
				goto parsing_failed;
			}
			if (parse_size(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(zero_time, string) == 0) {
			if (section != LIMIT_SECTION) {
				isnot_expected_msg("zero_time");
				goto parsing_failed;
			}
			if (REGEXEC(time_exp_val, sp) != 0) {
				wrong_format_msg(sp, "zero_time", "");
				goto parsing_failed;
			}
			if (parse_time_exp(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(expire_time, string) == 0) {
			if (section != EXPIRE_SECTION) {
				isnot_expected_msg("expire_time");
				goto parsing_failed;
			}
			if (REGEXEC(time_exp_val, sp) != 0) {
				wrong_format_msg(sp, "expire_time", "");
				goto parsing_failed;
			}
			if (parse_time_exp(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(exec, string) == 0) {
			if (section != REACH_SECTION && section != EXPIRE_SECTION &&
			    section != STARTUP_SECTION && section != IF_LIMIT_SECTION &&
			    section != IF_NOLIMIT_SECTION && section != SHUTDOWN_SECTION) {
				isnot_expected_msg("exec");
				goto parsing_failed;
			}
			if (parse_exec(string, sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(acl, string) == 0) {
			if (section != RULE_SECTION && section != GLOBAL_SECTION) {
				isnot_expected_msg("acl");
				goto parsing_failed;
			}
			if (REGEXEC(acl_val, sp) != 0) {
				wrong_format_msg(sp, "acl", "");
				goto parsing_failed;
			}
			if (parse_acl(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(maxchunk, string) == 0) {
			if (section != RULE_SECTION && section != GLOBAL_SECTION) {
				isnot_expected_msg("maxchunk");
				goto parsing_failed;
			}
			if (REGEXEC(size_val, sp) != 0) {
				wrong_format_msg(sp, "maxchunk", "");
				goto parsing_failed;
			}
			if (parse_size(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(info, string) == 0) {
			if (section != RULE_SECTION && section != LIMIT_SECTION) {
				isnot_expected_msg("info");
				goto parsing_failed;
			}
			if (parse_info(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(db_owner, string) == 0) {
			if (section != GLOBAL_SECTION) {
				isnot_expected_msg("db_owner");
				goto parsing_failed;
			}
			if (REGEXEC(db_owner_val, sp) != 0) {
				wrong_format_msg(sp, "db_owner", "");
				goto parsing_failed;
			}
			if (parse_db_owner(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(db_perm, string) == 0) {
			if (section != GLOBAL_SECTION) {
				isnot_expected_msg("db_perm");
				goto parsing_failed;
			}
			if (REGEXEC(db_perm_val, sp) != 0) {
				wrong_format_msg(sp, "db_perm", "");
				goto parsing_failed;
			}
			parse_db_perm(sp);	/* always correct value */
		} else if (REGEXEC(worktime, string) == 0) {
			if (section != GLOBAL_SECTION && section != RULE_SECTION &&
			    section != LIMIT_SECTION) {
				isnot_expected_msg("worktime");
				goto parsing_failed;
			}
			if (REGEXEC(worktime_val, sp) != 0) {
				wrong_format_msg(sp, "worktime", "");
				goto parsing_failed;
			}
			if (parse_worktime(sp) < 0)
				goto parsing_failed;
			if (mode != TEST_PARSING)
				Syslog(LOG_WARNING, "%s line %u: \"worktime\" parameter is not implemeted yet",
				    cfgfilename, lineno);
		} else if (REGEXEC(db_dir, string) == 0) {
			if (section != GLOBAL_SECTION) {
				isnot_expected_msg("db_dir");
				goto parsing_failed;
			}
			if (parse_db_dir(sp) < 0)
				goto parsing_failed;
		} else if (REGEXEC(debug_exec, string) == 0) {
			if (section != DEBUG_SECTION) {
				isnot_expected_msg("debug_exec");
				goto parsing_failed;
			}
			if (parse_debug_level(&debug_exec, sp) < 0)
				goto parsing_failed;
			debug_exec_is_set = 1;
		} else if (REGEXEC(debug_limit, string) == 0) {
			if (section != DEBUG_SECTION) {
				isnot_expected_msg("debug_limit");
				goto parsing_failed;
			}
			if (parse_debug_level(&debug_limit, sp) < 0)
				goto parsing_failed;
			debug_limit_is_set = 1;
		} else if (REGEXEC(debug_ipfw, string) == 0) {
			if (section != DEBUG_SECTION) {
				isnot_expected_msg("debug_ipfw");
				goto parsing_failed;
			}
#ifdef __FreeBSD__
			if (parse_debug_level(&debug_ipfw, sp) < 0)
				goto parsing_failed;
			debug_ipfw_is_set = 1;
#else
			line_err(string, "IP Firewall is not supported. Don't use \"debug_ipfw\" parameter.");
			goto parsing_failed;
#endif /* __FreeBSD__ */
		} else if (REGEXEC(debug_ipfil, string) == 0) {
			if (section != DEBUG_SECTION) {
				isnot_expected_msg("debug_ipfil");
				goto parsing_failed;
			}
			if (parse_debug_level(&debug_ipfil, sp) < 0)
				goto parsing_failed;
			debug_ipfil_is_set = 1;
		} else if (REGEXEC(debug_time, string) == 0) {
			if (section != DEBUG_SECTION) {
				isnot_expected_msg("debug_time");
				goto parsing_failed;
			}
			if (parse_debug_level(&debug_time, sp) < 0)
				goto parsing_failed;
			debug_time_is_set = 1;
		} else if (REGEXEC(debug_lock, string) == 0) {
			if (section != DEBUG_SECTION) {
				isnot_expected_msg("debug_lock");
				goto parsing_failed;
			}
			if (parse_debug_level(&debug_lock, sp) < 0)
				goto parsing_failed;
			debug_lock_is_set = 1;
		} else if (REGEXEC(lock_db, string) == 0) {
			if (section != GLOBAL_SECTION) {
				isnot_expected_msg("lock_db");
				goto parsing_failed;
			}
			if (REGEXEC(yesno_val, sp) != 0) {
				wrong_format_msg(sp, "lock_db", "");
				goto parsing_failed;
			}
			lock_db = parse_yesno(sp);
			lock_db_is_set = 1;
		} else if (REGEXEC(lock_wait_time, string) == 0) {
			if (section != GLOBAL_SECTION) {
				isnot_expected_msg("lock_wait_time");
				goto parsing_failed;
			}
			if (REGEXEC(time_val, sp) != 0) {
				wrong_format_msg(sp, "lock_wait_time", "");
				goto parsing_failed;
			}
			if (parse_lock_wait_time(sp) < 0)
				goto parsing_failed;
			if (mode != TEST_PARSING && lock_wait_time > 5 * MINUTE)
				Syslog(LOG_WARNING, "value for parameter \"lock_wait_time\" greater than 5 minutes isn't recommended");
		} else {
			line_err(string, "unknown parameter or syntax");
			goto parsing_failed;	
		}
	}
parsing_failed:
	Syslog(LOG_ERR, "can't parse configuration file");
	if (fclose(fp) != 0)
		Syslog(LOG_ERR, "fclose(%s): %m", cfgfilename);
	if (buf)
		free(buf);
	if (ac.rulename != NULL)
		free(ac.rulename);
	return -1;
}

/*
 * Build some regular expressions (regex). It helps to write much more
 * simple code for all parse_*() functions.
 */
static int
build_config_regexes(void)
{
#define pat_brace	"^}$"
#define	pat_rule	"^rule[ \t]+[^ \t]+[ \t]*{$"
#define pat_global	"^global[ \t]*{$"
#define pat_ipfw	"^ipfw[ \t]*="
#ifdef __FreeBSD__
#define pat_ipfw_val	"^-?[[:digit:]]+(\\.[[:digit:]]+)?([ \t]+-?[[:digit:]]+(\\.[[:digit:]]+)?)*$"
#endif
#define pat_ipfil	"^ipfil[ \t]*="
#define pat_ipfil_val	"^-?[io]([[:digit:]]+)?@[[:digit:]]+([ \t]+-?[io]([[:digit:]]+)?@[[:digit:]]+)*$"
#define	pat_update_db_time "^update_db_time[ \t]*="
#define pat_append_db_time "^append_db_time[ \t]*="
#define pat_time_val	"^([[:digit:]]+[smh][ \t]*)+$"
#define pat_emptyline	"^[ \t]*$"
#define	pat_size_val	"^([[:digit:]]+[bkmgtBKMGT][ \t]*)+$"
#define pat_byte_limit	"^bytes?_limit[ \t]*="
#define pat_acl		"^acl[ \t]*="
#define pat_limit	"^limit[ \t]+[^ \t]+[ \t]*{$"
#define pat_exec	"^exec[ \t]*(\\(:[-_[:alnum:]]+\\)|\\([-_[:alnum:]]+(:([-_[:alnum:]]+)?)?\\))?[ \t]*="
#define pat_expire	"^expire[ \t]*{$"
#define pat_zero_time	"^zero_time[ \t]*="
#define pat_expire_time	"^expire_time[ \t]*="
#define pat_time_exp_val "^(\
(\\+[mhdwDWM])?[ \t]*(([[:digit:]]+[smhdwDWM][ \t]*)+)?|\
(([[:digit:]]+[smhdwDWM][ \t]*)+)?[ \t]*(\\+[mhdwDWM])?)$"
#define pat_startup	"^startup[ \t]*{$"
#define pat_if_limit	"^(if_limit|if_limit_reached|if_limit_is_reached)[ \t]*{$"
#define pat_if_nolimit	"^(if_nolimit|if_limit_not_reached|if_limit_is_not_reached)[ \t]*{$"
#define pat_maxchunk	"^maxchunk[ \t]*="
#define pat_shutdown	"^shutdown[ \t]*{$"
#define pat_info	"^info[ \t]*="
#define pat_db_owner	"^db_owner[ \t]*="
#define pat_db_owner_val "^(:[-_[:alnum:]]+|[-_[:alnum:]]+(:([-_[:alnum:]]+)?)?)$"
#define pat_db_perm	"^db_perm[ \t]*="
#define pat_db_perm_val	"^(u|g|ug|gu)\\+r$"
#define pat_worktime	"^worktime[ \t]*="
#define pat_worktime_val "\
^[smtwhfaSMTWHFA][ \t]*(\\*|[[:digit:]]{1,2}:[[:digit:]]{1,2}-[[:digit:]]{1,2}:[[:digit:]]{1,2}([ \t]+[[:digit:]]{1,2}:[[:digit:]]{1,2}-[[:digit:]]{1,2}:[[:digit:]]{1,2})*)\
([ \t]+[smtwhfaSMTWHFA][ \t]*(\\*|[[:digit:]]{1,2}:[[:digit:]]{1,2}-[[:digit:]]{1,2}:[[:digit:]]{1,2}([ \t]+[[:digit:]]{1,2}:[[:digit:]]{1,2}-[[:digit:]]{1,2}:[[:digit:]]{1,2})*))*$"
#define pat_db_dir	"^db_dir[ \t]*="
#define pat_debug	"^debug[ \t]*{$"
#define pat_debug_ipfw	"^debug_ipfw[ \t]*="
#define pat_debug_ipfil	"^debug_ipfil[ \t]*="
#define pat_debug_limit	"^debug_limit[ \t]*="
#define pat_debug_exec	"^debug_exec[ \t]*="
#define pat_debug_time	"^debug_time[ \t]*="
#define pat_debug_lock	"^debug_lock[ \t]*="
#define pat_reach	"^reach[ \t]*{$"
#define pat_lock_db	"^lock_db[ \t]*="
#define pat_yesno_val	"^([yY][eE][sS]|[nN][oO])$"
#define pat_lock_wait_time "^lock_wait_time[ \t]*="

	static int already_built = 0;

#define	REGCOMP(x)							\
	if (regcomp(&reg_ ## x, pat_ ## x, REG_EXTENDED) != 0) {	\
		Syslog(LOG_ERR, "regcomp(" #x ")");			\
		return -1;						\
	}

	if (already_built == 1)
		return 0;
	REGCOMP(brace);
	REGCOMP(rule);
	REGCOMP(global);
	REGCOMP(ipfw);
#ifdef __FreeBSD__
	REGCOMP(ipfw_val);
#endif
	REGCOMP(update_db_time);
	REGCOMP(append_db_time);
	REGCOMP(emptyline);
	REGCOMP(byte_limit);
	REGCOMP(size_val);
	REGCOMP(time_val);
	REGCOMP(ipfil);
	REGCOMP(ipfil_val);
	REGCOMP(acl);
	REGCOMP(acl_val);
	REGCOMP(limit);
	REGCOMP(exec);
	REGCOMP(time_exp_val);
	REGCOMP(expire);
	REGCOMP(startup);
	REGCOMP(if_limit);
	REGCOMP(if_nolimit);
	REGCOMP(zero_time);
	REGCOMP(expire_time);
	REGCOMP(maxchunk);
	REGCOMP(shutdown);
	REGCOMP(info);
	REGCOMP(db_owner);
	REGCOMP(db_owner_val);
	REGCOMP(db_perm);
	REGCOMP(db_perm_val);
	REGCOMP(worktime);
	REGCOMP(worktime_val);
	REGCOMP(db_dir);
	REGCOMP(debug);
	REGCOMP(debug_ipfw);
	REGCOMP(debug_ipfil);
	REGCOMP(debug_limit);
	REGCOMP(debug_exec);
	REGCOMP(debug_time);
	REGCOMP(debug_lock);
	REGCOMP(reach);
	REGCOMP(lock_db);
	REGCOMP(yesno_val);
	REGCOMP(lock_wait_time);
	already_built = 1;
	return 0;
}

/*
 * Following functions are used for viewing configuration file, size, time, etc.
 * in human-readable format.
 */

/*
 * Convert size to human readable string (Tbytes, Gbytes, etc.).
 */
char *
show_size(u_quad_t a)
{
	char		*ptr;
	u_quad_t	t, g, m, k, b;

	t = a / TBYTE;
	a -= t * TBYTE;
	g = a / GBYTE;
	a -= g * GBYTE;
	m = a / MBYTE;
	a -= m * MBYTE;
	k = a / KBYTE;
	b = a - k * KBYTE;
	if (size_str != NULL)
		free(size_str);
	if (asprintf(&size_str, "%quT %quG %quM %quK %quB", t, g, m, k , b) < 0) {
		Syslog(LOG_ERR, "show_size:asprintf: %m");
		return "(error)";
	}
	ptr = size_str;
	if (t == 0) {
		ptr = skip_first_space(ptr);
		if (g == 0) {
			ptr = skip_first_space(ptr);
			if (m == 0) {
				ptr = skip_first_space(ptr);
				if (k == 0)
					ptr = skip_first_space(ptr);
			}
		}
	}
	if (b == 0) {
		last_space_to_nul(ptr);
		if (k == 0) {
			last_space_to_nul(ptr);
			if (m == 0) {
				last_space_to_nul(ptr);
				if (g == 0)
					last_space_to_nul(ptr);
			}
		}
	}
	return *ptr == '\0' ? "0B" : ptr;
}

/*
 * show_size2() is wrapper for show_size() and should be called when when
 * we want to call show_size() "two times".
 */
char *
show_size2(u_quad_t a)
{
	static char	*size_str2 = NULL;
	char		*size_str_save = size_str, *ptr;

	if (size_str2 != NULL)
		free(size_str2);
	size_str = NULL;
	ptr = show_size(a);
	size_str2 = size_str;
	size_str = size_str_save;
	return ptr;
}

/*
 * Convert time to human readable string (hours, minutes and seconds).
 */
char *
show_time(u_int a)
{
	static char	*str = NULL;
	char		*ptr;
	u_int		h, m, s;

	h = a / HOUR;
	a -= h * HOUR;
	m = a / MINUTE;
	s = a - m * MINUTE;
	if (str)
		free(str);
	if (asprintf(&str, "%uh %um %us", h, m, s) < 0) {
		Syslog(LOG_ERR, "show_time:asprintf: %m");
		return "(error)";
	}
	ptr = str;
	if (h == 0) {
		ptr = skip_first_space(ptr);
		if (m == 0)
			ptr = skip_first_space(ptr);
	}
	if (s == 0) {
		last_space_to_nul(ptr);
		if (m == 0)
			last_space_to_nul(ptr);
	}
	return *ptr == '\0' ? "0s" : ptr;
}

/*
 * Convert expired time to human readable sring.
 */
static char *
show_time_exp(struct time_exp *time_exp)
{
	static char	*str = NULL;
	char		*ptr;
	u_int		a = time_exp->seconds;
	u_int		d, h, m, s;

	d = a / DAY;
	a -= d * DAY;
	h = a / HOUR;
	a -= h * HOUR;
	m = a / MINUTE;
	s = a - m * MINUTE;
	if (str)
		free(str);
	if (asprintf(&str, "   %uM %uD %uh %um %us   ", time_exp->monthes, d, h, m, s) < 0) {
		Syslog(LOG_ERR, "show_time_exp:asprintf: %m");
		return "(error)";
	}
	*(str + strlen(str) - 3) = '\0';
	ptr = str + 3;
	if (time_exp->monthes == 0) {
		ptr = skip_first_space(ptr);
		if (d == 0) {
			ptr = skip_first_space(ptr);
			if (h == 0) {
				ptr = skip_first_space(ptr);
				if (m == 0)
					ptr = skip_first_space(ptr);
			}
		}
	}
	if (s == 0) {
		last_space_to_nul(ptr);
		if (m == 0) {
			last_space_to_nul(ptr);
			if (h == 0) {
				last_space_to_nul(ptr);
				if (d == 0)
					last_space_to_nul(ptr);
			}
		}
	}
	if (time_exp->upto != UPTO_SIMPLE) {
		if (time_exp->side == 0) {
			ptr -= 3;
			*ptr = '+';
			*(ptr + 1) = time_exp->upto;
		} else {
			size_t	len = strlen(ptr);
			*(ptr + len) = ' ';
			*(ptr + len + 1) = '+';
			*(ptr + len + 2) = time_exp->upto;
			/* *(ptr + len - 3) = '\0'; -- isn't necessary */
		}
	}
	return *ptr == '\0' ? "0s" : ptr;
}

/*
 * Output "exec" parameter.
 */
static void
show_exec(struct cmd *cmdp)
{
	struct passwd	*passwdp;
	struct group	*groupp;

	printf("exec");
	if (cmdp->uid_is_set || cmdp->gid_is_set) {
		if (cmdp->uid_is_set) {
			if (cmdp->uid_is_named) {
				if ( (passwdp = getpwuid(cmdp->uid)) == NULL)
					printf("(%lu", cmdp->uid);
				else
					printf("(%s", passwdp->pw_name);
			} else
				printf("(%lu", cmdp->uid);
		} else
			printf("(");
		if (cmdp->gid_is_set) {
			if  (cmdp->gid_is_named) {
				if ( (groupp = getgrgid(cmdp->gid)) == NULL)
					printf(":%lu)", cmdp->gid);
				else
					printf(":%s)", groupp->gr_name);
			} else
				printf(":%lu)", cmdp->gid);
		} else
			printf(":)");
	}
	printf(" = %s\n", cmdp->str);
}

/*
 * Output "startup" and "shutdown" sections with "exec" parameters.
 * Also output "if_limit_is_reached" and "if_limit_is_not_reached" sections
 * if they are present.
 */
static void
show_commands(struct commands *commandsp, const char *firstline, const char *prefix)
{
	struct cmd	*cmdp;
	u_int		k;
	
	if (commandsp->cmd || commandsp->cmd_if_limit || commandsp->cmd_if_nolimit) {
		printf("%s%s {\n", prefix, firstline);
		for (k = 0, cmdp = commandsp->cmd; k < commandsp->ncmd; ++cmdp, ++k) {
			printf("%s    ", prefix);
			show_exec(cmdp);
		}
		if (commandsp->cmd_if_limit) {
			printf("%s    if_limit_is_reached {\n", prefix);
			for (k = 0, cmdp = commandsp->cmd_if_limit; k < commandsp->ncmd_if_limit; ++cmdp, ++k) {
				printf("%s        ", prefix);
				show_exec(cmdp);
			}
			printf("%s    }\n", prefix);
		}
		if (commandsp->cmd_if_nolimit) {
			printf("%s    if_limit_is_not_reached {\n", prefix);
			for (k = 0, cmdp = commandsp->cmd_if_nolimit; k < commandsp->ncmd_if_nolimit; ++cmdp, ++k) {
				printf("%s        ", prefix);
				show_exec(cmdp);
			}
			printf("%s    }\n", prefix);
		}
		printf("%s}\n", prefix);
	}
}

/*
 * Output "db_owner" parameter.
 */
static void
show_db_owner(void)
{
	struct passwd	*pwd;
	struct group	*grp;

	if (db_uid_is_set || db_gid_is_set) {
		printf("    db_owner = ");
		if (db_uid_is_set) {
			if (db_uid_is_named) {
				if ( (pwd = getpwuid(db_uid)) == NULL)
					printf("%lu", db_uid);
				else
					printf("%s", pwd->pw_name);
			} else
				printf("%lu", db_uid);
		}
		if (db_gid_is_set) {
			if (db_gid_is_named) {
				if ( (grp = getgrgid(db_gid)) == NULL)
					printf(":%lu\n", db_gid);
				else
					printf(":%s\n", grp->gr_name);
			} else
				printf(":%lu\n", db_gid);
		} else
			printf(":\n");
	}
}

/*
 * Output "db_perm" parameter.
 */
static void
show_db_perm(void)
{
	if (db_perm_is_set)
		printf("    db_perm = %s\n",
		    db_fileperm == FILE_PERM_RUSR ? "u+r" : (
		    db_fileperm == FILE_PERM_RGRP ? "g+r" :
						    "ug+r"
		));
}

/*
 * Output "worktime" parameter.
 */
static void
show_worktime(struct worktime *wtp, const char *prefix)
{
	u_int		wday, i;
	char		*days = "SMTWHFA";
	struct interval	*intp;

	printf("%sworktime =", prefix);
	for (wday = 0; wday < 7; ++wday)
		if (wtp[wday].interval != NULL) {
			printf(" %c", days[wday]);
			for (i = 0, intp = wtp[wday].interval; i < wtp[wday].ninterval; ++intp, ++i)
				if (intp->h2 == 24)
					printf(" *");
				else
					printf(" %02d:%02d-%02d:%02d",
					    intp->h1, intp->m1,
					    intp->h2, intp->m2);
		}
	printf("\n");
}

/*
 * Main function for outputing indented configuration file.
 */
void
show_config(void)
{
	u_int		i, j, k;
	char		ipfilac_type;
	u_short		ipfilac_group;
	struct ipac	*acp;
#ifdef __FreeBSD__
	struct ipfwac	*ipfwacp;
#endif
	struct ipfilac	*ipfilacp;
	struct ipfilac_list *list;
	struct limit	*lp;
	struct cmd	*cmdp;

	if (global_section_is_set) {
		printf("global {\n");
		if (update_db_time_global)
			printf("    update_db_time = %s\n", show_time(update_db_time_global));
		if (append_db_time_global != 0)
			printf("    append_db_time = %s\n", show_time(append_db_time_global));
		if (acl_global)
			printf("    acl = %s\n", acl_global);
		if (maxchunk_global != 0)
			printf("    maxchunk = %s\n", show_size(maxchunk_global));
		show_db_owner();
		show_db_perm();
		if (db_dir_is_set)
			printf("    db_dir = %s\n", db_dir);
		if (use_worktime_global)
			show_worktime(worktime_global, "    ");
		if (lock_db_is_set)
			printf("    lock_db = %s\n", lock_db ? "yes" : "no");
		if (lock_wait_time_is_set)
			printf("    lock_wait_time = %s\n", show_time(lock_wait_time));
		printf("}\n\n");
	}
	if (debug_section_is_set) {
		printf("debug {\n");
#ifdef __FreeBSD__
		if (debug_ipfw_is_set)
			printf("    debug_ipfw = %d\n", debug_ipfw);
#endif
		if (debug_ipfil_is_set)
			printf("    debug_ipfil = %d\n", debug_ipfil);
		if (debug_exec_is_set)
			printf("    debug_exec = %d\n", debug_exec);
		if (debug_limit_is_set)
			printf("    debug_limit = %d\n", debug_limit);
		if (debug_time_is_set)
			printf("    debug_time = %d\n", debug_time);
		if (debug_lock_is_set)
			printf("    debug_lock = %d\n", debug_lock);
		printf("}\n\n");
	}
	if (startup_global.cmd) {
		show_commands(&startup_global, "startup", "");
		printf("\n");
	}
	if (shutdown_global.cmd) {
		show_commands(&shutdown_global, "shutdown", "");
		printf("\n");
	}
	for (i = 0, acp = ipac; i < nipac; ++i, ++acp) {
		printf("rule %s {\n", acp->rulename);
		if (acp->update_db_time != 0)
			printf("    update_db_time = %s\n", show_time(acp->update_db_time));
		if (acp->append_db_time != 0)
			printf("    append_db_time = %s\n", show_time(acp->append_db_time));
#ifdef __FreeBSD__
		if (acp->ipfwac) {
			printf("    ipfw =");
			for (j = 0, ipfwacp = acp->ipfwac; j < acp->nipfwac; ++ipfwacp, ++j)
				printf("%s%hu.%u", ipfwacp->action < 0 ? " -" : " ", ipfwacp->number, ipfwacp->subnumber);
			printf("\n");
		}
#endif
		for (j = 0, ipfilacp = acp->ipfilac; j < acp->nipfilac; ++ipfilacp, ++j) {
			ipfilac_type = ipfilacp->type;
			ipfilac_group = ipfilacp->group;
			printf("    ipfil =");
			for (k = 0, list = ipfilacp->list; k < ipfilacp->nlist; ++list, ++k)
				printf("%s%c%hu@%u", list->action < 0 ? " -" : " ",
				     ipfilac_type, ipfilac_group, list->number);
			printf("\n");
		}
		if (acp->acl)
			printf("    acl = %s\n", acp->acl);
		if (acp->maxchunk != 0)
			printf("    maxchunk = %s\n", show_size(acp->maxchunk));
		if (acp->use_worktime)
			show_worktime(acp->worktime, "    ");
		show_commands(&acp->rc[0], "startup", "    ");
		show_commands(&acp->rc[1], "shutdown", "    ");
		for (j = 0, lp = acp->limit; j < acp->nlimit; ++lp, ++j) {
			printf("    limit %s {\n", lp->limitname);
			printf("        byte_limit = %s\n", show_size(lp->byte_limit));
			if (lp->zero_time_param.upto != UPTO_NOTSET)
				printf("        zero_time = %s\n", show_time_exp(&lp->zero_time_param));
			if (lp->cmd != NULL) {
				printf("        reach {\n");
				for (k = 0, cmdp = lp->cmd; k < lp->ncmd; ++cmdp, ++k) {
					printf("            ");
					show_exec(cmdp);
				}
				printf("        }\n");
			}
			if (lp->use_worktime)
				show_worktime(lp->worktime, "        ");
			if (lp->expire.time.upto != UPTO_NOTSET) {
				printf("        expire {\n");
				printf("            expire_time = %s\n", show_time_exp(&lp->expire.time));
				for (k = 0, cmdp = lp->expire.cmd; k < lp->expire.ncmd; ++cmdp, ++k) {
					printf("            ");
					show_exec(cmdp);
				}
				printf("        }\n");
			}
			show_commands(&lp->rc[0], "startup", "        ");
			show_commands(&lp->rc[1], "shutdown", "        ");
			printf("    }\n");
		}
		printf("}\n\n");
	}
}
