/*-
 * Copyright (c) 2000 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: db.c,v 1.1.2.4 2001/01/02 11:55:28 simon Exp $";
#endif /* !lint */

#include <errno.h>
#include <fcntl.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>
#include <varargs.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <regex.h>

#include "db.h"

#include "ipa.h"
#include "common.h"
#include "config.h"
#include "debug.h"
#include "path.h"
#include "rules.h"

/* PERMMASK allows to get just permission bits from st_mode field
   PERMMAS(x) = x & 0777 */
#define PERMMASK(x)	((x) & (S_IRWXU|S_IRWXG|S_IRWXO))

char		*db_dir = DBDIR;		/* "db_dir" parameter, main database directory */
u_long		db_uid, db_gid;			/* "db_owner" parameter */
u_int		db_fileperm, db_dirperm;	/* "db_perm" parameter */
int		lock_db;			/* "lock_db" parameter */
u_int		lock_wait_time;			/* "lock_wait_time" parameter */
time_t		curr_time;			/* current UTC time in seconds */
struct tm	curr_tm;			/* current localtime */

regex_t		reg_db_entry, reg_limit_line1, reg_limit_timestamp;

static int	db_fd;
static char	*lock_db_file;
static int	db_file_is_locked = 0;
static jmp_buf	env_alrm;

static struct sigaction sig_alrm2_act, save_sig_alrm;

static int	check_db_deep(char *);
static int	check_db_dir_ownership(const char *, struct stat *);
static int	chown_db(const char *);
static int	init_db_rules(void);

/*
 * Yet another ALRM signal handler.
 * Should be used to avoid race condition in "lock" functions.
 */
static void
sig_alrm2(int signo)
{
	longjmp(env_alrm, 1);
}

static void
init_sig_alrm2(void)
{
	sig_alrm2_act.sa_flags = 0;
	sigemptyset(&sig_alrm2_act.sa_mask);
	sigaddset(&sig_alrm2_act.sa_mask, SIGHUP);
	sigaddset(&sig_alrm2_act.sa_mask, SIGTERM);
	sigaddset(&sig_alrm2_act.sa_mask, SIGCHLD);
	if (debug)
		sigaddset(&sig_alrm2_act.sa_mask, SIGINT);
	sig_alrm2_act.sa_handler = sig_alrm2;
}

static int
set_sig_alrm2(void)
{
	if (sigaction(SIGALRM, &sig_alrm2_act, &save_sig_alrm) < 0) {
		syslog(LOG_ERR, "sigaction(SIGALRM), set new handler: %m");
		return -1;
	}
	return 0;
}

static int
reset_sig_alrm(void)
{
	if (sigaction(SIGALRM, &save_sig_alrm, (struct sigaction *)0) < 0) {
		syslog(LOG_ERR, "sigaction(SIGALRM), restore handler: %m");
		return -1;
	}
	return 0;
}

/*
 * Obtain "write" lock on file from starting offset 0 bytes
 * until the end of file. Starting offset is measured from the position
 * specified by whence parameter (SEEK_CUR, SEEK_SET and SEEK_END).
 */
int
lock_db_file_until_end(char *file, int fd, short whence)
{
	if (lock_db > 0) {
		if (lock_wait_time > 0) { /* "lock_wait_time" parameter is used */
			if (set_sig_alrm2() < 0)
				return -1;
			if (setjmp(env_alrm) != 0) {
				syslog(LOG_WARNING, "Can't lock file %s during %u seconds, give up.",
				    file, lock_wait_time);
				db_file_is_locked = 0;
				if (reset_sig_alrm() < 0)
					return -1;
				return 0;
			}
			if (debug_lock > 0)
				syslog(LOG_INFO, "lock_db_file_until_end: try to lock file %s (timeout %u seconds)",
				    file, lock_wait_time);
			alarm(lock_wait_time);
			if (writew_lock(fd, 0, whence, 0) < 0) {
				syslog(LOG_ERR, "fcntl(%s, F_SETLKW): %m", file);
				return -1;
			}
			alarm(0);
			if (debug_lock > 0)
				syslog(LOG_INFO, "lock_db_file_until_end: obtained lock on file %s",
				    file);
			if (reset_sig_alrm() < 0)
				return -1;
		} else {
			if (debug_lock > 0)
				syslog(LOG_INFO, "lock_db_file_until_end: try to lock file %s", file);
			if (writew_lock(fd, 0, whence, 0) < 0) {
				syslog(LOG_ERR, "fcntl(%s, F_SETLKW): %m", file);
				return -1;
			}
			if (debug_lock > 0)
				syslog(LOG_INFO, "lock_db_file_until_end: obtained lock on file %s", file);
		}
		db_file_is_locked = 1;
	}
	return 0;
}

/*
 * Unlock whole file. At any time only one file is locked, that why
 * we can use db_file_is_locked variable here.
 */
int
unlock_db_file(char *file, int fd)
{
	if (db_file_is_locked > 0) {
		if (debug_lock > 0)
			syslog(LOG_INFO, "unlock_db_file: release lock on %s (FD %d)", file, fd);
		if (un_lock(fd, 0, SEEK_SET, 0) < 0) {
			syslog(LOG_ERR, "unlock_db_file: fcntl(%s, F_UNLCK): %m", file);
			return -1;
		}
		db_file_is_locked = 0;
	}
	return 0;
}

/*
 * Obtain exclusive lock on whole database.
 */
static int
lock_db_write(void)
{
	if (lock_db > 0) {
		if ( (db_fd = open(lock_db_file, O_WRONLY)) < 0) {
			syslog(LOG_ERR, "open(%s, O_WRONLY): %m", lock_db_file);
			return -1;
		}
		if (lock_db_file_until_end(lock_db_file, db_fd, SEEK_SET) < 0)
			return -1;
	}
	return 0;
}

/*
 * Unlock whole database.
 */
static int
unlock_db(void)
{
	if (db_file_is_locked > 0) {
		if (debug_lock > 0)
			syslog(LOG_INFO, "unlock_db: release lock on %s (FD %d)", lock_db_file, db_fd);
		if (close(db_fd) < 0) {
			syslog(LOG_ERR, "close(%s): %m", lock_db_file);
			return -1;
		}
		db_file_is_locked = 0;
	}
	return 0;
}

/*
 * Update current record for *acp accounting rule in database.
 * value is equal to new value of counter.
 */
int
update_db(struct ipac *acp, u_quad_t value)
{
	if (fseek(acp->fp, (long)-DBCOUNTEROFF, SEEK_END) < 0) {
		syslog(LOG_ERR, "fseek(%s, -%d, SEEK_END): %m",
		    acp->filename, DBCOUNTEROFF);
		return -1;
	}

	if (lock_db_file_until_end(acp->filename, acp->fd, SEEK_CUR) < 0)
		return 0;

	/* update second timestamp and counter */
	if (fprintf(acp->fp, "%02d:%02d:%02d %020qu",
	    curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec, value) != DBCOUNTERSIZE) {
		syslog(LOG_ERR, "update_db: fprintf(%s), failed: %m", acp->filename);
		return -1;
	}
	if (fflush(acp->fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", acp->filename);
		return -1;
	}
	
	return unlock_db_file(acp->filename, acp->fd);
}

/*
 * Append new record for *acp accounting rule in database.
 * value is equal to value of new record.
 */
int
append_db(struct ipac *acp, u_quad_t value)
{
	if (fseek(acp->fp, 0L, SEEK_END) < 0) {
		syslog(LOG_ERR, "fseek(%s, 0, SEEK_END): %m", acp->filename);
		return -1;
	}

	if (lock_db_file_until_end(acp->filename, acp->fd, SEEK_CUR) < 0)
		return -1;

	if (fprintf(acp->fp, "%02d/%02d:%02d:%02d-%02d:%02d:%02d %020qu\n",
	    curr_tm.tm_mday,
	    curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec,
	    curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec,
	    value) != DBRECORDSIZE) {
		syslog(LOG_ERR, "append_db: fprintf(%s), failed: %m", acp->filename);
		return -1;
	}
	/* newrec_time will be checked if "append_db_time" parameter was
	   specified in the configuration file for *acp rule */
	acp->newrec_time = curr_time + acp->append_db_time;
	if (fflush(acp->fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", acp->filename);
		return -1;
	}
	return unlock_db_file(acp->filename, acp->fd);
}

/*
 * Check ownership and mode of main database directory. If ownership or/and
 * mode is/are not equal to proper values, then test whole database files and
 * directories.
 */
static int 
check_db_ownership(void)
{
	struct stat	statbuf;

	if (stat(db_dir, &statbuf) < 0) {
		syslog(LOG_ERR, "stat(%s): %m", db_dir);
		return -1;
	}
	if (statbuf.st_uid != db_uid || statbuf.st_gid != db_gid ||
	    PERMMASK(statbuf.st_mode) != db_dirperm) {
		syslog(LOG_WARNING, "database main directory %s has wrong permissions or has wrong owner", db_dir);
		syslog(LOG_INFO, "checking all database...");
		if (check_db_dir_ownership(db_dir, &statbuf) < 0)
			return -1;
		if (check_db_deep(db_dir) < 0)
			return -1;
	}
	return 0;
}

/*
 * Create directory in database and set proper owner of new directory.
 */
int
mkdir_db(const char *path)
{
	if (mkdir(path, db_dirperm) < 0) {
		syslog(LOG_ERR, "mkdir(%s, 0%03o): %m", path, db_dirperm);
		return -1;
	}
	return chown_db(path);
}

int
init_db(void)
{
	struct stat	statbuf;

	init_sig_alrm2();

	if (stat(db_dir, &statbuf) < 0) {
		if (errno != ENOENT) {
			syslog(LOG_ERR, "stat(%s): %m", db_dir);
			return -1;
		}
		syslog(LOG_INFO, "creating database main directory %s", db_dir);
		if (mkdir_db(db_dir) < 0)
			return -1;
	}

	if (asprintf(&lock_db_file, "%s/"LOCKDBFILE, db_dir) < 0) {
		syslog(LOG_ERR, "init_db: asprintf: %m");
		return -1;
	}

	if (stat(lock_db_file, &statbuf) < 0) {
		int	fd;

		if (errno != ENOENT) {
			syslog(LOG_ERR, "stat(%s): %m", lock_db_file);
			return -1;
		}
		syslog(LOG_INFO, "creating database lock file %s", lock_db_file);
		if ( (fd = open(lock_db_file, O_WRONLY|O_CREAT, db_fileperm)) < 0) {
			syslog(LOG_ERR, "open(%s, O_WRONLY|O_CREAT, 0%03o): %m", lock_db_file, db_fileperm);
			return -1;
		}
		if (close(fd) < 0) {
			syslog(LOG_ERR, "init_db: close(%s): %m", lock_db_file);
			return -1;
		}
	}

	if (lock_db_write() < 0) {
		syslog(LOG_ERR, "Can't lock database %s", db_dir);
		return -1;
	}

	if (check_db_ownership() < 0)
		return -1;

	if (init_db_rules() < 0)
		return -1;

	time(&curr_time);
	localtime_r(&curr_time, &curr_tm);
	if (init_limits(1) < 0)
		return -1;

	if (unlock_db() < 0) {
		syslog(LOG_ERR, "Can't unlock database %s", db_dir);
		return -1;
	}
	return 0;
}

FILE *
create_db_file(const char *path)
{
	int	fd;
	FILE	*fp;

	if ( (fd = open(path, O_RDWR|O_CREAT, db_fileperm)) < 0) {
		syslog(LOG_ERR, "open(%s, O_RDWR|O_CREAT, 0%03o): %m", path, db_fileperm);
		return NULL;
	}
	if (chown_db(path) < 0)
		return NULL;
	/* some file can be already exists, truncate it */
	if (ftruncate(fd, (off_t)0) < 0) {
		syslog(LOG_ERR, "ftruncate(%s, 0): %m", path);
		return NULL;
	}
	if ( (fp = fdopen(fd, "r+")) == NULL)
		syslog(LOG_ERR, "fdopen(%s, \"r+\"): %m", path);
	return fp;
}

int
check_db_dir_ownership(const char *path, struct stat *statbuf)
{
	if (!S_ISDIR(statbuf->st_mode)) {
		syslog(LOG_ERR, "%s should be directory", path);
		return -1;
	}
	if (statbuf->st_uid != db_uid || statbuf->st_gid != db_gid) {
		syslog(LOG_WARNING, "directory %s is owned by UID %u GID %u",
		    path, statbuf->st_uid, statbuf->st_gid);
		syslog(LOG_WARNING, "changing ownership of directory %s to UID %lu GID %lu", path, db_uid, db_gid);
		if (chown_db(path) < 0)
			return -1;
	}
	if (PERMMASK(statbuf->st_mode) != db_dirperm) {
		syslog(LOG_WARNING, "directory %s has permissions 0%03o",
		    path, PERMMASK(statbuf->st_mode));
		syslog(LOG_WARNING, "changing permissions for directory %s to 0%03o",
		    path, db_dirperm);
		if (chmod(path, db_dirperm) < 0) {
			syslog(LOG_ERR, "chmod(%s, 0%03o): %m", path, db_dirperm);
			return -1;
		}
	}
	return 0;
}

int
check_db_file_ownership(const char *path, struct stat *statbuf)
{
	if (!S_ISREG(statbuf->st_mode)) {
		syslog(LOG_ERR, "%s should be regular file", path);
		return -1;
	}
	if (statbuf->st_uid != db_uid || statbuf->st_gid != db_gid) {
		syslog(LOG_WARNING, "file %s is owned by UID %u GID %u",
		    path, statbuf->st_uid, statbuf->st_gid);
		syslog(LOG_WARNING, "changing owner of file %s to UID %lu GID %lu",
		    path, db_uid, db_gid);
		if (chown_db(path) < 0)
			return -1;
	}
	if (PERMMASK(statbuf->st_mode) != db_fileperm) {
		syslog(LOG_WARNING, "file %s has permissions 0%03o",
		    path, PERMMASK(statbuf->st_mode));
		syslog(LOG_WARNING, "changing permissions for file %s to 0%03o",
		    path, db_fileperm);
		if (chmod(path, db_fileperm) < 0) {
			syslog(LOG_ERR, "chmod(%s, 0%03o): %m", path, db_fileperm);
			return -1;
		}
	}
	return 0;
}

/*
 * Recursive function for checking owner and mode of database files
 * and directories.
 */
static int
check_db_deep(char *start)
{
	char		*path;
	DIR		*dp;
	struct stat	statbuf;
	struct dirent	*dirp;

	if ( (dp = opendir(start)) == NULL) {
		syslog(LOG_ERR, "opendir(%s): %m", start);
		return -1;
	}
	while ( (dirp = readdir(dp)) != 0) {
		if (strcmp(dirp->d_name, ".") == 0 || strcmp(dirp->d_name, "..") == 0)
			continue;
		if (asprintf(&path, "%s/%s", start, dirp->d_name) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			syslog(LOG_ERR, "stat(%s): %m", path);
			return -1;
		}
		if (S_ISREG(statbuf.st_mode)) {
			if (check_db_file_ownership(path, &statbuf) < 0)
				return -1;
		} else if (S_ISDIR(statbuf.st_mode)) {
			if (check_db_dir_ownership(path, &statbuf) < 0)
				return -1;
			if (check_db_deep(path) < 0)
				return -1;
		} else {
			syslog(LOG_ERR, "%s must be a regular file or directory", path);
			return -1;
		}
		free(path);
	}
	if (closedir(dp) < 0) {
		syslog(LOG_ERR, "closedir(%s): %m", start);
		return -1;
	}
	return 0;
}

/*
 * Set proper owner of database directory or file.
 */
int
chown_db(const char *path)
{
	if (chown(path, db_uid, db_gid) < 0) {
		syslog(LOG_ERR, "chown(%s, %lu, %lu): %m", path, db_uid, db_gid);
		return -1;
	}
	return 0;
}

/*
 * Init accouting rules: recreate info, acl files, create/open limits.
 * XXX any file shouldn't be locked, because whole database is locked in
 * init_db() function.
 */
int
init_db_rules(void)
{
	char		*path, *acl;
	u_int		i, j;
	FILE		*fp;
	struct ipac	*acp;
	struct limit	*lp;
	struct stat	statbuf;

	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		/* check directory for rule */
		if (asprintf(&path, "%s/%s", db_dir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			if (errno != ENOENT)
				syslog(LOG_WARNING, "stat(%s): %m", path);
			syslog(LOG_INFO, "creating rule directory %s", path);
			if (mkdir_db(path) < 0)
				return -1;
		}
		free(path);

		/* check directory for limits */
		if (asprintf(&path, "%s/%s/" LIMITSDIR, db_dir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			if (errno != ENOENT)
				syslog(LOG_WARNING, "stat(%s): %m", path);
			if (mkdir_db(path) < 0)
				return -1;
		}
		free(path);

		/* create/overwrite acl file */
		if (asprintf(&path, "%s/%s/" ACLFILE, db_dir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if ( (acl = acp->acl ? acp->acl : acl_global) != NULL) {
			if ( (fp = create_db_file(path)) == NULL)
				return -1;
			if (fprintf(fp, "%s", acl) != strlen(acl)) {
				syslog(LOG_ERR, "fprintf(%s): %m", path);
				return -1;
			}
			if (fclose(fp) != 0) {
				syslog(LOG_ERR, "fclose(%s): %m", path);
				return -1;
			}
			free(acp->acl);
			acp->acl = NULL;
		} else {
			if (unlink(path) < 0)
				if (errno != ENOENT) {
					syslog(LOG_ERR, "unlink(%s): %m", path);
					return -1;
				}
		}
		free(path);

		/* create/owerwrite info file */
		if (asprintf(&path, "%s/%s/" INFOFILE, db_dir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (acp->info) {
			if ( (fp = create_db_file(path)) == NULL)
				return -1;
			if (fprintf(fp, "%s", acp->info) != strlen(acp->info)) {
				syslog(LOG_ERR, "fprintf(%s): %m", path);
				return -1;
			}
			if (fclose(fp) != 0) {
				syslog(LOG_ERR, "fclose(%s): %m", path);
				return -1;
			}
			free(acp->info);
			acp->info = NULL;
		} else {
			if (unlink(path) < 0)
				if (errno != ENOENT) {
					syslog(LOG_ERR, "unlink(%s): %m", path);
					return -1;
				}
		}
		free(path);

		/* create directories for limits */
		for (j = 0, lp = acp->limit; j < acp->nlimit; ++j, ++lp) {
			/* create directory for limit */
			if (asprintf(&path, "%s/%s/" LIMITSDIR "/%s", db_dir,
			    acp->rulename, lp->limitname) < 0) {
				syslog(LOG_ERR, "asprintf: %m");
				return -1;
			}
			if (stat(path, &statbuf) < 0) {
				if (errno != ENOENT)
					syslog(LOG_ERR, "stat(%s): %m", path);
				syslog(LOG_INFO, "creating directory for rule %s limit %s",
				    acp->rulename, lp->limitname);
				if (mkdir_db(path) < 0)
					return -1;
			}
			free(path);

			/* create or open file for limit accounting */
			if (asprintf(&path, "%s/%s/" LIMITSDIR "/%s/" LIMITFILE,
			    db_dir, acp->rulename, lp->limitname) < 0) {
				syslog(LOG_ERR, "asprintf: %m");
				return -1;
			}
			if (stat(path, &statbuf) < 0) {
				if (errno != ENOENT)
					syslog(LOG_WARNING, "stat(%s): %m", path);
				if ( (lp->fp = create_db_file(path)) == NULL)
					return -1;
			} else {
				if (statbuf.st_size == 0)
					syslog(LOG_WARNING, "file %s is empty, it will be fixed.", path);
				if ( (lp->fp = fopen(path, "r+")) == NULL) {
					syslog(LOG_ERR, "fopen(%s, \"r+\"): %m", path);
					return -1;
				}
			}
			lp->fd = fileno(lp->fp);
			lp->filename = path;

			/* create info file for limit */
			if (asprintf(&path, "%s/%s/" LIMITSDIR "/%s/" INFOFILE,
			    db_dir, acp->rulename, lp->limitname) < 0) {
				syslog(LOG_ERR, "asprintf: %m");
				return -1;
			}
			if (lp->info) {
				if ( (fp = create_db_file(path)) == NULL)
					return -1;
				if (fprintf(fp, "%s", lp->info) != strlen(lp->info)) {
					syslog(LOG_ERR, "fprintf(%s): %m", path);
					return -1;
				}
				if (fclose(fp) != 0) {
					syslog(LOG_ERR, "fclose(%s): %m", path);
					return -1;
				}
				free(lp->info);
				lp->info = NULL;
			} else {
				if (unlink(path) < 0)
					if (errno != ENOENT) {
						syslog(LOG_ERR, "unlink(%s): %m", path);
						return -1;
					}
			}
			free(path);
		}
	}
	if (acl_global) {
		free(acl_global);
		acl_global = NULL;
	}
	return 0;
}

/*
 * Build RE to check database correct values.
 */
int
build_db_regexes(void)
{
#define	REGCOMP(x)							\
	if (regcomp(&reg_ ## x, pat_ ## x, REG_EXTENDED) != 0) {	\
		syslog(LOG_ERR, "regcomp(" #x ")");			\
		return -1;						\
	}

	REGCOMP(db_entry);
	REGCOMP(limit_line1);
	REGCOMP(limit_timestamp);
	return 0;
}

/*
 * Just send information about number of loaded accounting rules to syslog.
 */
void
loaded_rules_info(void)
{
	if (nipac == 0)
		syslog(LOG_WARNING, "can't find any rule in configuration file, hope that's OK");
	else
		syslog(LOG_INFO, "loaded %u accounting rule%s", nipac, nipac == 1 ? "" : "s");
}
