/*-
 * Copyright (c) 2000 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: ipastat.c,v 1.1.2.10 2001/01/02 11:55:28 simon Exp $";
#endif /* !lint */

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <setjmp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <grp.h>
#include <pwd.h>
#include <regex.h>

#include "ipastat.h"

#include "common.h"
#include "path.h"


enum { Full = 0, Kbyte, Mbyte, Gbyte, Tbyte };
static char	*bcnt_convert_msg[] = { "TGMKbytes", "Kbytes", "Mbytes", "Gbytes", "Tbytes" };

static int	bcnt_convert_flag = Mbyte; /* by default convert counters to Mbyte */
static char	*rule = NULL, *limit = NULL, *time_interval = NULL;
static char	*db_dir = DBDIR;	/* database directory */
static int	ac_per_day_flag = 1,	/* output accounting per day or per timestamp */
		set_ids_flag = 0,	/* == 1 if UID == 0 or -u switch is set */
		use_second_timestamp_flag = 0, /* == 1 if -I switch is set */
		check_dates_flag = 1,	/* == 0 if -e switch is set */
		print_table_flag = 1;	/* == 0 if -n whitch is set */
static int	*gid_list = NULL;	/* GID and supplementary GIDs */
static u_int	ngid_list = 0;

static uid_t	my_uid;	/* my UID */
static gid_t	my_gid; /* my GID */

static struct tm	curr_tm;

static int	db_fd;
static jmp_buf	env_alrm;
static char	*lockdbfile = LOCKDBFILE;
static char	*lockdbpath;

#define LOCK_WAIT_TIME	10	/* max #seconds wait for lock */


#ifdef __GNUC__
static void	err_exit(const char *, ...) __attribute__ ((format (printf, 1, 2)));
static void	warn_exit(const char *, ...) __attribute__ ((format (printf, 1, 2)));
static void	format_err(const char *, ...) __attribute__ ((format (printf, 1, 2)));
#else
static void	err_exit(const char *, ...);
static void	warn_exit(const char *, ...);
static void	format_err(const char *, ...);
#endif

static void	show_usage(char *), show_version(void);
static int	readline(char **, size_t *, FILE *, char *);
static void	get_my_ids(void);
static int	fullstrtoul(u_long *, const char *);
static char	* get_info(char *, char *);
static char	* total_convert_bytes(u_quad_t);
static void	lock_db_read(void);
static void	unlock_db(void);
static void	set_sighandlers(void);
static void	lock_db_file_until_end(char *, int, int, short);

/*
 * Regular expressions for checking time interval given in the -i or -I option.
 */
#define pat_time_interval_pre "^[^-]+(-[^-]+)?$"
#define pat_time_interval "\
^(\
(\
[[:digit:]]{4,}(\\.[[:digit:]]{1,2})?(\\.[[:digit:]]{1,2})?|\
([[:digit:]]{1,2})?(\\.[[:digit:]]{1,2})?\
)?\
(\
/[[:digit:]]{1,2}(:[[:digit:]]{1,2})?(:[[:digit:]]{1,2})?\
)?\
)$"

#define	REGCOMP(x)							\
	if (regcomp(&reg_ ## x, pat_ ## x, REG_EXTENDED) != 0)		\
		err_exit("regcomp(" #x ")");

static char	*no_rule_msg1 = "rule \"%s\" wasn't found.",
		*no_rule_msg2 = "you aren't allowed to browse rule \"%s\" or this rule wasn't found.";

struct name_list {
	char	*name;
	char	*info;
};

static char	use_e_switch_msg[] = "(use -e switch to view information)";


/*
 * Check if given rule in command line is present in database.
 * If given rule doesn't exists, then display warning message and exit.
 */
static void
check_rule(void)
{
	char		*path;
	struct stat	statbuf;

	if (strcmp(rule, lockdbfile) == 0)
		warn_exit(no_rule_msg1, rule);
	if (asprintf(&path, "%s/%s", db_dir, rule) < 0)
		err_exit("asprintf");
	if (stat(path, &statbuf) < 0) {
		if (errno != ENOENT)
			err_exit("stat(%s)", path);
		if (my_uid == 0)
			warn_exit(no_rule_msg1, rule);
		else
			warn_exit(no_rule_msg2, rule);
	}
	free(path);
}

/*
 * Output error message about errors in time interval specified
 * in the -i or -I option.
 */
static void
time_interval_err(const char *message, int side)
{
	format_err("time-interval (wrong %s, %s side)", message, side == 0 ? "left" : "right");
}

/*
 * Parse left or right side of time interval specified in the -i or -I option.
 * Format of time interval is described in ipastat(8).
 * Complete left or right part if it is needed.
 */
static void
parse_time_interval(char *s, struct tm *tm_ptr, int side)
{
	int		n, val1, val2, val3;
	char		*ptr;

	memcpy(tm_ptr, &curr_tm, sizeof *tm_ptr);
	/* initial value for time in right part,
	   for left part these fields are completed below */
	tm_ptr->tm_hour = tm_ptr->tm_min = tm_ptr->tm_sec = 0;

	if (*s != '/') {
		/* date is specified */
		if (*s == '.') {
			/* .DD */
			if (sscanf(++s, "%d", &tm_ptr->tm_mday) != 1)
				err_exit("sscanf(%%d): failed");
		} else {
			if ( (n = sscanf(s, "%d.%d.%d", &val1, &val2, &val3)) < 1)
				err_exit("sscanf(%%d.%%d.%%d): failed");
			if ( (ptr = strchr(s, '.')) == NULL)
				if ( (ptr = strchr(s, '/')) == NULL)
					ptr = strchr(s, '\0');
			if (n == 3) {
				/* YYYY.MM.DD */
				tm_ptr->tm_year = val1;
				tm_ptr->tm_mon = val2;
				tm_ptr->tm_mday = val3;
			} else if (n == 2) {
				if (ptr - s >= 4) {
					/* YYYY.MM */
					tm_ptr->tm_year = val1;
					tm_ptr->tm_mon = val2;
					tm_ptr->tm_mday = side == 0 ? 1 : 31;
				} else {
					/* MM.DD */
					tm_ptr->tm_mon = val1;
					tm_ptr->tm_mday = val2;
				}
			} else /* n == 1 */ {
				if (ptr - s >= 4) {
					/* YYYY */
					tm_ptr->tm_year = val1;
					if (side == 0)
						tm_ptr->tm_mon = tm_ptr->tm_mday = 1;
					else {
						tm_ptr->tm_mon = 12;
						tm_ptr->tm_mday = 31;
					}
				} else {
					/* MM */
					tm_ptr->tm_mon = val1;
					tm_ptr->tm_mday = side == 0 ? 1 : 31;
				}
			}
		}
	}

	n = 0;
	if ( (s = strchr(s, '/')) != NULL) {
		/* time is specified */
		if ( (n = sscanf(++s, "%d:%d:%d", &val1, &val2, &val3)) < 1)
			err_exit("sscanf(%%d:%%d:%%d): failed");
		if (n >= 1)
			tm_ptr->tm_hour = val1;	/* hh */
		if (n >= 2)
			tm_ptr->tm_min = val2;  /* hh:mm */
		if (n == 3)
			tm_ptr->tm_sec = val3;  /* hh:mm:ss */
	}

	if (tm_ptr->tm_mon == 0 || tm_ptr->tm_mon > 12)
		time_interval_err("month", side);
	if (tm_ptr->tm_mday == 0 || tm_ptr->tm_mday > 31)
		time_interval_err("day", side);
	if (tm_ptr->tm_hour > 23)
		time_interval_err("hours", side);
	if (tm_ptr->tm_min > 59)
		time_interval_err("minutes", side);
	if (tm_ptr->tm_sec > 59)
		time_interval_err("seconds", side);

	if (side == 1)	/* for left part complete value of time if it is needed */
		switch (n) {
		case 2:
			if (++tm_ptr->tm_min > 59)
				tm_ptr->tm_min = 0;
			else
				break;
		case 1:
			if (++tm_ptr->tm_hour > 23) {
				tm_ptr->tm_hour = 0;
			} else
				break;
		case 0:
			if (++tm_ptr->tm_mday > 31) {
				tm_ptr->tm_mday = 1;
				if (++tm_ptr->tm_mon > 12) {
					tm_ptr->tm_mon = 1;
					++tm_ptr->tm_year;
				}
			}
		}
}

/*
 * Create buffer and save human readable date and time in it.
 */
static char *
tm_str(struct tm *tm_ptr)
{
	char	*buf;

	if (asprintf(&buf, "%d.%02d.%02d/%02d:%02d:%02d",
	    tm_ptr->tm_year, tm_ptr->tm_mon, tm_ptr->tm_mday,
	    tm_ptr->tm_hour, tm_ptr->tm_min, tm_ptr->tm_sec) < 0)
		err_exit("asprintf");
	return buf;
}

/*
 * Create buffer and save converted value in it.
 */
static char *
convert_bytes(u_quad_t value)
{
	char		*buf;
	u_quad_t	conv_value;

	switch (bcnt_convert_flag) {
	case Kbyte:
		conv_value = value / KBYTE;
		break;
	case Mbyte:
		conv_value = value / MBYTE;
		break;
	case Gbyte:
		conv_value = value / GBYTE;
		break;
	case Tbyte:
		conv_value = value / TBYTE;
		break;
	default: /* Full */
		return total_convert_bytes(value);
	}
	if (asprintf(&buf, "%qu", conv_value) < 0)
		err_exit("aspintf");
	return buf;
}

/*
 * Output one line of table.
 * n - number of arguments,
 * ... - each argument is unsigned integer equal to width of column.
 */
static void
print_table_line(u_int n, ...)
{
	va_list	ap;
	u_int	width, i;

	if (!print_table_flag)
		return;
	va_start(ap, n);
	while (n--) {
		printf("+");
		width = va_arg(ap, u_int);
		for (i = 0; i < width + 2; ++i)
			printf("-");
	}
	printf("+\n");
}

/*
 * Check if user is allowed to view accounting information for rule rulename.
 * Return 0 if user isn't allowed, else return 1.
 */
static int
acl_is_allowed(char *rulename)
{
	static char	*buf = NULL;
	static size_t	bufsize;
	static int	already_built = 0;
	static regex_t	reg_acl_val;

	int		len, is_allowed_flag;
	char		*s, *name_start, char_save;
	int		*gidp;
	u_int		i;
	u_long		val;
	char		*path;
	FILE		*fp;
	struct passwd	*passwdp;
	struct group	*groupp;


	if (set_ids_flag != 0)
		return 1; /* don't use ACL if -u switch is used */
	if (asprintf(&path, "%s/%s/" ACLFILE, db_dir, rulename) < 0)
		err_exit("asprintf");
	if ( (fp = fopen(path, "r")) == NULL)
		if (errno != ENOENT && errno != EACCES)	
			err_exit("fopen(%s, \"r\")", path);
	if (!fp) {
		free(path);
		return 0;
	}
	len = readline(&buf, &bufsize, fp, path);
	if (len == 0) {
		if (fclose(fp) != 0)
			err_exit("fclose(%s)", path);
		free(path);
		return 0;
	}

	s = skip_spaces(buf);
	remove_trailing_spaces(s);

	if (already_built == 0) {
		REGCOMP(acl_val);
		already_built = 1;
	}
	if (REGEXEC(acl_val, s) != 0)
		format_err("file %s", path);

	for (;;) {
		s = skip_spaces(s);
		if (*s == '\0') {
			is_allowed_flag = 0;
			break;
		}
		if (*s == '!') {
			++s;
			is_allowed_flag = 0;
		} else
			is_allowed_flag = 1;
		name_start = s;
		s = skip_chars(s);
		char_save = *s;
		*s = '\0';
		if (*name_start == '%') {
			++name_start;
			if (strcmp(name_start, "ALL%") == 0)
				break;
			if ( (groupp = getgrnam(name_start)) == NULL) {
				if (fullstrtoul(&val, name_start) < 0) {
					*s = char_save;
					continue;
				}
			} else
				val = groupp->gr_gid;
			for (gidp = gid_list, i = 0; i < ngid_list; ++gidp, ++i)
				if (*gidp == val)
					break;
		} else {
			if ( (passwdp = getpwnam(name_start)) == NULL) {
				if (fullstrtoul(&val, name_start) < 0) {
					*s = char_save;
					continue;
				}
			} else
				val = passwdp->pw_uid;
			if (my_uid == val)
				break;
		}
		*s = char_save;
	}
	if (fclose(fp) != 0)
		err_exit("fclose(%s)", path);
	free(path);
	return is_allowed_flag;
}

/*
 * Output accounting information for rule.
 */
static void
print_rule_ac(void)
{
	char		oneline[DBRECORDSIZE + 1];
	char		*split_ptr = NULL, *info, *path;
	int		year, mon, ret, prev_mday;
	u_int		u_quad_max_summary_cnt = 0, u_quad_max_cnt_per_day = 0,
			nrecords = 0, ndays = 0, i;
	char		*u_quad_max_str, *u_quad_max_conv_str = NULL; /* initial value isn't used */
	regex_t		reg_db_entry, reg_time_interval_pre, reg_time_interval;
	u_quad_t	db_value, summary = 0, summary_per_day = 0;
	char		*summary_str, *summary_conv_str;
	FILE		*fp;
	char		*tm1_ptr, *tm2_ptr;
	u_int		width_1_1, width_1_2, width_1_3, width_1_4,
			width_2_1, width_2_2, width_2_3, width_2_4 = 0, width_2_5 = 0,
			tmp;
	struct tm	tm1, tm2, db_tm1, db_tm2;
	struct stat	statbuf;
	struct ac_list {
		char	*bytes;
		char	*bytes_conv;
		char	*date;
		u_int	u_quad_max_cnt;
		u_short	h1, m1, s1, h2, m2, s2;
	} * ac_list = NULL;
	u_int		nac_list = 0;
	char		*format;
	time_t		curr_time;
	u_int		lineno;
	
	time(&curr_time);
	localtime_r(&curr_time, &curr_tm);
	++curr_tm.tm_mon;
	curr_tm.tm_year += 1900;
	if (time_interval == NULL) {
		/* current month */
		memcpy(&tm1, &curr_tm, sizeof tm1);
		memcpy(&tm2, &tm1, sizeof tm2);
		tm1.tm_hour = tm1.tm_min = tm1.tm_sec =
		    tm2.tm_hour = tm2.tm_min = tm2.tm_sec = 0;
		tm1.tm_mday = tm2.tm_mday = 1;
		if (++tm2.tm_mon > 12) {
			tm2.tm_mon = 1;
			++tm2.tm_year;
		}
	} else {
		REGCOMP(time_interval_pre);
		REGCOMP(time_interval);
		if (REGEXEC(time_interval_pre, time_interval) != 0)
			format_err("time-interval (wrong parts)");
		if ( (split_ptr = strchr(time_interval, '-')) != NULL) {
			*split_ptr = '\0';
			if (REGEXEC(time_interval, time_interval) != 0)
				format_err("time-interval (wrong left part)");
			++split_ptr;
			if (REGEXEC(time_interval, split_ptr) != 0)
				format_err("time-interval (wrong right part)");
		} else {
			if (REGEXEC(time_interval, time_interval) != 0)
				format_err("time-interval");
			split_ptr = time_interval;
			memcpy(&tm2, &tm1, sizeof tm2);
		}
		parse_time_interval(time_interval, &tm1, 0);
		parse_time_interval(split_ptr, &tm2, 1);
	}

	REGCOMP(db_entry);
	if (tmcmp(&tm1, &tm2) > 0)
		format_err("time-interval (date in right part should be greater than date in left part)");

	check_rule();

	if (acl_is_allowed(rule) == 0)
		warn_exit(no_rule_msg2, rule);

	info = get_info(rule, (char *)0);

	width_1_1 = 4; /* Rule */
	width_1_2 = 4; /* Info */
	width_1_3 = 4; /* From */
	width_1_4 = 2; /* To */

	if ( (tmp = strlen(rule)) > width_1_1)
		width_1_1 = tmp;
	if (info)
		if ( (tmp = strlen(info)) > width_1_2)
			width_1_2 = tmp;
	tm1_ptr = tm_str(&tm1);
	tm2_ptr = tm_str(&tm2);
	if ( (tmp = strlen(tm1_ptr)) > width_1_3)
		width_1_3 = tmp;
	if ( (tmp = strlen(tm2_ptr)) > width_1_4)
		width_1_4 = tmp;
	
	width_2_1 = 4;	/* Date */
	if (ac_per_day_flag) {
		width_2_2 = 5; /* Bytes */
		width_2_3 = strlen(bcnt_convert_msg[bcnt_convert_flag]);
	} else {
		width_2_2 = 8; /* hh:mm:ss */
		width_2_3 = 8; /* hh:mm:ss */
		width_2_4 = 5; /* Bytes */
		width_2_5 = strlen(bcnt_convert_msg[bcnt_convert_flag]);
	}
	for (year = tm1.tm_year; year <= tm2.tm_year; ++year) {
		int	mon1, mon2;

		mon1 = year == tm1.tm_year ? tm1.tm_mon : 1;
		mon2 = year == tm2.tm_year ? tm2.tm_mon : 12;
		for (mon = mon1; mon <= mon2; ++mon) {
			if (asprintf(&path, "%s/%s/%d/%02d", db_dir, rule, year, mon) < 0)
				err_exit("asprintf");
			if (stat(path, &statbuf) < 0) {
				if (errno != ENOENT && errno != EACCES)
					err_exit("stat(%s)", path);
				free(path);
				continue;
			}
			if (!S_ISREG(statbuf.st_mode))
				err_exit("%s should be regular file", path);
			if ( (fp = fopen(path, "r")) == NULL) {
				if (errno != ENOENT && errno != EACCES)
					err_exit("fopen(%s, \"r\")", path);
				free(path);
				continue;
			}
			db_tm1.tm_year = db_tm2.tm_year = year;
			db_tm1.tm_mon = db_tm2.tm_mon = mon;

			if (fstat(fileno(fp), &statbuf) < 0)
				err_exit("fstat(%s)", path);
			if (statbuf.st_size < DBRECORDSIZE)
				format_err("file %s (file is too small)", path);
			
			lock_db_file_until_end(path, fileno(fp), -DBRECORDSIZE, SEEK_END);

			for (prev_mday = 0, lineno = 0;;) {
				struct tm	prev_db_tm2;

				if ( (ret = fread(oneline, sizeof(char), DBRECORDSIZE, fp)) != DBRECORDSIZE) {
					if (ret < DBRECORDSIZE && ret > 0) {
						if (feof(fp) != 0)
							format_err("file %s (can't recognize last line: file is too small)", path);
						err_exit("fread(%s): failed", path);
					}
					if (ret == 0) {
						if (feof(fp) != 0) /* EOF */
							break;
						else
							err_exit("fread(%s)", path);
					}
				}
				++lineno;
				oneline[DBRECORDSIZE] = '\0';
				if (REGEXEC(db_entry, oneline) != 0)
					format_err("file %s", path);
				if (sscanf(oneline, "%d/%d:%d:%d-%d:%d:%d %qu",
				    &db_tm1.tm_mday,
				    &db_tm1.tm_hour, &db_tm1.tm_min, &db_tm1.tm_sec,
				    &db_tm2.tm_hour, &db_tm2.tm_min, &db_tm2.tm_sec,
				    &db_value) != 8)
					err_exit("sscanf(%s, %%d/%%d:%%d:%%d-%%d:%%d:%%d %%qu): failed", path);
				db_tm2.tm_mday = db_tm1.tm_mday;
				if (check_dates_flag) {
					if (check_date(&db_tm1) < 0 ||
					    check_date(&db_tm2) < 0)
						format_err("file %s (line %u): wrong value of timestamp %s", path, lineno, use_e_switch_msg);
					if (tmcmp(&db_tm1, &db_tm2) > 0 ||
					    prev_mday > db_tm1.tm_mday)
						format_err("file %s (line %u): wrong sequence of dates %s", path, lineno, use_e_switch_msg);
					if (prev_mday == db_tm1.tm_mday)
						if (tmcmp(&prev_db_tm2, &db_tm1) > 0)
							format_err("file %s (line %u): wrong sequence of dates %s", path, lineno, use_e_switch_msg);
					prev_db_tm2 = db_tm2;
				}
				if (use_second_timestamp_flag && tmcmp(&db_tm2, &tm2) > 0)
					continue;
				if (tmcmp(&db_tm1, &tm1) < 0 || tmcmp(&db_tm1, &tm2) >= 0)
					continue;
				if (ac_per_day_flag) {
					if (prev_mday == 0) {
						++ndays;
						prev_mday = db_tm1.tm_mday;
						summary_per_day = db_value;
						u_quad_max_cnt_per_day = 0;
					} else if (prev_mday != db_tm1.tm_mday) {
						++ndays;
						if ( (ac_list = realloc(ac_list, (nac_list + 1) * sizeof *ac_list)) == NULL)
							err_exit("realloc");
						if (asprintf(&ac_list[nac_list].date, "%d.%02d.%02d", year, mon, prev_mday) < 0)
							err_exit("asprintf");
						if ( (tmp = strlen(ac_list[nac_list].date)) > width_2_1)
							width_2_1 = tmp;
						if (asprintf(&ac_list[nac_list].bytes, "%qu", summary_per_day) < 0)
							err_exit("asprintf");
						if ( (tmp = strlen(ac_list[nac_list].bytes)) > width_2_2)
							width_2_2 = tmp;
						ac_list[nac_list].bytes_conv = convert_bytes(summary_per_day);
						if ( (tmp = strlen(ac_list[nac_list].bytes_conv)) > width_2_3)
							width_2_3 = tmp;
						ac_list[nac_list].u_quad_max_cnt = u_quad_max_cnt_per_day;
						++nac_list;
						prev_mday = db_tm1.tm_mday;
						summary_per_day = db_value;
					} else {
						if (summary_per_day < UQUAD_MAX - db_value)
							summary_per_day += db_value;
						else {
							++u_quad_max_cnt_per_day;
							summary_per_day = UQUAD_MAX - summary_per_day;
							summary_per_day = db_value - summary_per_day;
						}
					}
				} else {
					++nrecords;
					if (prev_mday == 0) {
						++ndays;
						prev_mday = db_tm1.tm_mday;
					} else if (prev_mday != db_tm1.tm_mday) {
						++ndays;
						prev_mday = db_tm1.tm_mday;
					}
					if ( (ac_list = realloc(ac_list, (nac_list + 1) * sizeof *ac_list)) == NULL)
						err_exit("realloc");
					if (asprintf(&ac_list[nac_list].date, "%d.%02d.%02d", year, mon, prev_mday) < 0)
						err_exit("asprintf");
					if ( (tmp = strlen(ac_list[nac_list].date)) > width_2_1)
						width_2_1 = tmp;
					if (asprintf(&ac_list[nac_list].bytes, "%qu", db_value) < 0)
						err_exit("asprintf");
					if ( (tmp = strlen(ac_list[nac_list].bytes)) > width_2_4)
						width_2_4 = tmp;
					ac_list[nac_list].bytes_conv = convert_bytes(db_value);
					if ( (tmp = strlen(ac_list[nac_list].bytes_conv)) > width_2_5)
						width_2_5 = tmp;
					ac_list[nac_list].u_quad_max_cnt = u_quad_max_cnt_per_day;
					ac_list[nac_list].h1 = db_tm1.tm_hour;
					ac_list[nac_list].m1 = db_tm1.tm_min;
					ac_list[nac_list].s1 = db_tm1.tm_sec;
					ac_list[nac_list].h2 = db_tm2.tm_hour;
					ac_list[nac_list].m2 = db_tm2.tm_min;
					ac_list[nac_list].s2 = db_tm2.tm_sec;
					++nac_list;
				}
				if (summary < UQUAD_MAX - db_value)
					summary += db_value;
				else {
					++u_quad_max_summary_cnt;
					summary = UQUAD_MAX - summary;
					summary = db_value - summary;
				}
			}
			if (fclose(fp) != 0) /* also release lock */
				err_exit("fclose(%s)", path);
			free(path);
			if (ac_per_day_flag && ndays > 0 && prev_mday != 0) {
				if ( (ac_list = realloc(ac_list, (nac_list + 1) * sizeof *ac_list)) == NULL)
					err_exit("realloc");
				if (asprintf(&ac_list[nac_list].date, "%d.%02d.%02d", year, mon, prev_mday) < 0)
					err_exit("asprintf");
				if ( (tmp = strlen(ac_list[nac_list].date)) > width_2_1)
					width_2_1 = tmp;
				if (asprintf(&ac_list[nac_list].bytes, "%qu", summary_per_day) < 0)
					err_exit("asprintf");
				if ( (tmp = strlen(ac_list[nac_list].bytes)) > width_2_2)
					width_2_2 = tmp;
				ac_list[nac_list].bytes_conv = convert_bytes(summary_per_day);
				if ( (tmp = strlen(ac_list[nac_list].bytes_conv)) > width_2_3)
					width_2_3 = tmp;
				ac_list[nac_list].u_quad_max_cnt = u_quad_max_cnt_per_day; 
				++nac_list;
			}
		}
	}

	unlock_db();

	if (asprintf(&summary_str, "%qu", summary) < 0)
		err_exit("asprintf");
	summary_conv_str = convert_bytes(summary);
	if (ac_per_day_flag) {
		if ( (tmp = strlen(summary_str)) > width_2_2)
			width_2_2 = tmp;
		if ( (tmp = strlen(summary_conv_str)) > width_2_3)
			width_2_3 = tmp;
	} else {
		if ( (tmp = strlen(summary_str)) > width_2_4)
			width_2_4 = tmp;
		if ( (tmp = strlen(summary_conv_str)) > width_2_5)
			width_2_5 = tmp;
	}
	if (u_quad_max_summary_cnt != 0) {
		if (asprintf(&u_quad_max_str, "%qu", UQUAD_MAX) < 0)
			err_exit("asprintf");
		u_quad_max_conv_str = convert_bytes(UQUAD_MAX);
		if (ac_per_day_flag) {
			if ( (tmp = strlen(u_quad_max_str)) > width_2_2)
				width_2_2 = tmp;
			if ( (tmp = strlen(u_quad_max_conv_str)) > width_2_3)
				width_2_3 = tmp;
		} else {
			if ( (tmp = strlen(u_quad_max_str)) > width_2_4)
				width_2_4 = tmp;
			if ( (tmp = strlen(u_quad_max_conv_str)) > width_2_5)
				width_2_5 = tmp;
		}
	}

	format = "| %-*s | %-*s | %-*s | %-*s |\n";
	print_table_line(4, width_1_1, width_1_2, width_1_3, width_1_4);
	if (print_table_flag)
		printf(format, width_1_1, "Rule", width_1_2, "Info", width_1_3, "From", width_1_4, "To");
	print_table_line(4, width_1_1, width_1_2, width_1_3, width_1_4);
	if (print_table_flag)
		printf(format,width_1_1, rule, width_1_2, info ? info : "", width_1_3, tm1_ptr, width_1_4, tm2_ptr);
	else
		printf("%s %s %s %s\n", rule, info ? info : "", tm1_ptr, tm2_ptr);
	print_table_line(4, width_1_1, width_1_2, width_1_3, width_1_4);
	if (print_table_flag)
		printf(" \n");

	if (ac_per_day_flag) {
		format = "| %-*s | %*s | %*s |\n";
		print_table_line(3, width_2_1, width_2_2, width_2_3);
		if (print_table_flag)
			printf(format, width_2_1, "Date", width_2_2, "Bytes", width_2_3, bcnt_convert_msg[bcnt_convert_flag]);
		print_table_line(3, width_2_1, width_2_2, width_2_3);
		for (i = 0; i < nac_list; ++i) {
			if (ac_list[i].u_quad_max_cnt) {
				if (print_table_flag)
					printf(format, width_2_1, ac_list[i].date, width_2_2, u_quad_max_str, width_2_3, u_quad_max_conv_str);
				else
					printf("%s %s %s\n", ac_list[i].date, u_quad_max_str, u_quad_max_conv_str);
				while (--ac_list[i].u_quad_max_cnt) {
					if (print_table_flag)
						printf(format, width_2_1, "", width_2_2, u_quad_max_str, width_2_3, u_quad_max_conv_str);
					else
						printf("%s %s\n", u_quad_max_str, u_quad_max_conv_str);
				}
				if (print_table_flag)
					printf(format, width_2_1, "", width_2_2, ac_list[i].bytes, width_2_3, ac_list[i].bytes_conv);
				else
					printf("%s %s\n", ac_list[i].bytes, ac_list[i].bytes_conv);
			} else {
				if (print_table_flag)
					printf(format, width_2_1, ac_list[i].date, width_2_2, ac_list[i].bytes, width_2_3, ac_list[i].bytes_conv);
				else
					printf("%s %s %s\n", ac_list[i].date, ac_list[i].bytes, ac_list[i].bytes_conv);
			}
		}
		if (ac_list)
			print_table_line(3, width_2_1, width_2_2, width_2_3);
		while (u_quad_max_summary_cnt--) {
			if (print_table_flag)
				printf(format, width_2_1, "", width_2_2, u_quad_max_str, width_2_3, u_quad_max_conv_str);
			else
				printf("%s %s\n", u_quad_max_str, u_quad_max_conv_str);
		}
		if (print_table_flag)
			printf(format, width_2_1, "", width_2_2, summary_str, width_2_3, summary_conv_str);
		else
			printf("%s %s\n", summary_str,  summary_conv_str);
		print_table_line(3, width_2_1, width_2_2, width_2_3);
		if (print_table_flag)
			printf(" * %u day%s *\n", ndays,
			    ndays == 1 ? "" : "s");
	} else {
		format = "| %-*s | %-*s | %-*s | %*s | %*s |\n";
		print_table_line(5, width_2_1, width_2_2, width_2_3, width_2_4, width_2_5);
		if (print_table_flag)
			printf(format, width_2_1, "Date", width_2_2, "From", width_2_3, "To", width_2_4, "Bytes", width_2_5, bcnt_convert_msg[bcnt_convert_flag]);
		print_table_line(5, width_2_1, width_2_2, width_2_3, width_2_4, width_2_5);
		for (i = 0; i < nac_list; ++i) {
			if (print_table_flag)
				printf("| %-*s | %02d:%02d:%02d | %02d:%02d:%02d | %*s | %*s |\n",
				    width_2_1, ac_list[i].date,
				    ac_list[i].h1, ac_list[i].m1, ac_list[i].s1,
				    ac_list[i].h2, ac_list[i].m2, ac_list[i].s2,
				    width_2_4, ac_list[i].bytes, width_2_5, ac_list[i].bytes_conv);
			else
				printf("%s %02d:%02d:%02d %02d:%02d:%02d %s %s\n",
				    ac_list[i].date,
				    ac_list[i].h1, ac_list[i].m1, ac_list[i].s1,
				    ac_list[i].h2, ac_list[i].m2, ac_list[i].s2,
				    ac_list[i].bytes, ac_list[i].bytes_conv);
		}
		if (ac_list)
			print_table_line(5, width_2_1, width_2_2, width_2_3, width_2_4, width_2_5);
		while (u_quad_max_summary_cnt--) {
			if (print_table_flag)
				printf(format, width_2_1, "", width_2_2, "", width_2_3, "", width_2_4, u_quad_max_str, width_2_5, u_quad_max_conv_str);
			else
				printf("%s %s\n", u_quad_max_str, u_quad_max_conv_str);
		}
		if (print_table_flag)
			printf(format, width_2_1, "", width_2_2, "", width_2_3, "", width_2_4, summary_str, width_2_5, summary_conv_str);
		else
			printf("%s %s\n", summary_str,  summary_conv_str);
		print_table_line(5, width_2_1, width_2_2, width_2_3, width_2_4, width_2_5);
		if (print_table_flag)
			printf("* %u record%s, %u day%s *\n", nrecords,
			    nrecords == 1 ? "" : "s", ndays, ndays == 1 ? "" : "s");
	}
}

/*
 * Read data from info file.
 * If limitname is NULL then read info for rule rulename.
 */
static char *
get_info(char *rulename, char *limitname)
{
	int		len = 0;
	char		*path;
	FILE		*fp;
	struct stat	statbuf;
	static char	*buf = NULL;
	static size_t	bufsize;

	if (limitname == NULL) {
		if (asprintf(&path, "%s/%s/" INFOFILE, db_dir, rulename) < 0)
			err_exit("asprintf");
	} else {
		if (asprintf(&path, "%s/%s/limits/%s/" INFOFILE, db_dir, rulename, limitname) < 0)
			err_exit("asprintf");
	}
	if (stat(path, &statbuf) < 0) {
		if (errno != ENOENT && errno != EACCES)
			err_exit("stat(%s)", path);
		free(path);
		return NULL;
	}
	if (!S_ISREG(statbuf.st_mode))
		err_exit("%s should be regular file", path);
	if ( (fp = fopen(path, "r")) == NULL) {
		if (errno != ENOENT && errno != EACCES)
			err_exit("fopen(%s, \"r\")", path);
	} else {
		len = readline(&buf, &bufsize, fp, path);
		if (fclose(fp) != 0)
			err_exit("fclose(%s)", path);
	}
	free(path);
	return len == 0 ? NULL : buf;
}

/*
 * Compare two fields named name in struct name_list.
 * This function is called from qsort().
 */
static int
cmp_name_list(const void *p1, const void *p2)
{
	return strcmp(((struct name_list *)p1)->name, ((struct name_list *)p2)->name);
}

/*
 * Output all rules, which are allowed for user.
 */
static void
show_all_rules(void)
{
	DIR		*dp;
	char		*info;
	struct name_list *rules_list = NULL;
	u_int		nrules_list = 0;
	u_int		width_1, width_2, tmp, i;
	char		format[] = "| %-*s | %-*s |\n";
	struct dirent	*dirp;

	if ( (dp = opendir(db_dir)) == NULL)
		err_exit("opendir(%s)", db_dir);
	while ( (dirp = readdir(dp)) != NULL) {
		if (strcmp(dirp->d_name, ".") == 0 || strcmp(dirp->d_name, "..") == 0 ||
		    strcmp(dirp->d_name, LOCKDBFILE) == 0)
			continue;
		if (acl_is_allowed(dirp->d_name) == 0)
			continue;
		info = get_info(dirp->d_name, (char *)0);
		if ( (rules_list = realloc(rules_list, (nrules_list + 1) * sizeof *rules_list)) == NULL)
			err_exit("realloc");
		if (asprintf(&rules_list[nrules_list].name, "%s", dirp->d_name) < 0)
			err_exit("asprintf");
		if (info != NULL) {
			if (asprintf(&rules_list[nrules_list].info, "%s", info) < 0)
				err_exit("asprintf");
		} else
			rules_list[nrules_list].info = NULL;
		++nrules_list;
	}
	if (closedir(dp) < 0)
		err_exit("closedir(%s)", db_dir);
	if (nrules_list == 0) {
		if (my_uid != 0)
			warn_exit("you aren't allowed to view any rule, or no rules were found.");
		else
			warn_exit("no rules were found.");
	}

	unlock_db();

	/* Sort list by rules names */
	qsort(rules_list, nrules_list, sizeof *rules_list, cmp_name_list);

	width_1 = 4; /* Rule */
	width_2 = 4; /* Info */
	for (i = 0; i < nrules_list; ++i) {
		tmp = strlen(rules_list[i].name);
		if (width_1 < tmp)
			width_1 = tmp;
		if (rules_list[i].info != NULL) {
			tmp = strlen(rules_list[i].info);
			if (width_2 < tmp)
				width_2 = tmp;
		}
	}
	print_table_line(2, width_1, width_2);
	if (print_table_flag)
		printf(format, width_1, "Rule", width_2, "Info");
	print_table_line(2, width_1, width_2);
	for (i = 0; i < nrules_list; ++i) {
		if (print_table_flag)
			printf(format, width_1, rules_list[i].name, width_2, rules_list[i].info ? rules_list[i].info : "");
		else
			printf("%s %s\n", rules_list[i].name, rules_list[i].info ? rules_list[i].info : "");
	}
	print_table_line(2, width_1, width_2);
}

/*
 * Output all limits for rule, which are allowed for user.
 */
static void
show_all_limits(void)
{
	DIR		*dp;
	char		*info, *limits_dir;
	struct name_list *limits_list = NULL;
	u_int		nlimits_list = 0;
	struct dirent	*dirp;
	u_int		width_1_1, width_1_2, width_2_1, width_2_2, tmp, i;
	char		format[] = "| %-*s | %-*s |\n";

	check_rule();

	if (acl_is_allowed(rule) == 0)
		warn_exit(no_rule_msg2, rule);

	if (asprintf(&limits_dir, "%s/%s/limits", db_dir, rule) == NULL)
		err_exit("asprintf");
	if ( (dp = opendir(limits_dir)) == NULL)
		err_exit("opendir(%s)", limits_dir);

	while ( (dirp = readdir(dp)) != NULL) {
		if (strcmp(dirp->d_name, ".") == 0 || strcmp(dirp->d_name, "..") == 0)
			continue;
		info = get_info(rule, dirp->d_name);
		if ( (limits_list = realloc(limits_list, (nlimits_list + 1) * sizeof *limits_list)) == NULL)
			err_exit("realloc");
		if (asprintf(&limits_list[nlimits_list].name, "%s", dirp->d_name) < 0)
			err_exit("asprintf");
		if (info != NULL) {
			if (asprintf(&limits_list[nlimits_list].info, "%s", info) < 0)
				err_exit("asprintf");
		} else
			limits_list[nlimits_list].info = NULL;
		++nlimits_list;
	}
	if (closedir(dp) < 0)
		err_exit("closedir(%s)", db_dir);
	if (limits_list == 0)
		warn_exit("no limits were found in rule \"%s\".", rule);

	unlock_db();

	width_1_1 = 4; /* Rule */
	width_1_2 = 4; /* Info */
	tmp = strlen(rule);
	if (width_1_1 < tmp)
		width_1_1 = tmp;
	if ( (info = get_info(rule, (char *)0)) != NULL) {
		tmp = strlen(info);
		if (width_1_2 < tmp)
			width_1_2 = tmp;
	}

	width_2_1 = 5; /* Limit */
	width_2_2 = 4; /* Info */
	for (i = 0; i < nlimits_list; ++i) {
		tmp = strlen(limits_list[i].name);
		if (width_2_1 < tmp)
			width_2_1 = tmp;
		if (limits_list[i].info != NULL) {
			tmp = strlen(limits_list[i].info);
			if (width_2_2 < tmp)
				width_2_2 = tmp;
		}
	}

	/* Sort list by limits names */
	qsort(limits_list, nlimits_list, sizeof *limits_list, cmp_name_list);

	print_table_line(2, width_1_1, width_1_2);
	if (print_table_flag)
		printf(format, width_1_1, "Rule", width_1_2, "Info");
	print_table_line(2, width_1_1, width_1_2);
	if (print_table_flag)
		printf(format, width_1_1, rule, width_1_2, info ? info : "");
	else
		printf("%s %s\n", rule, info ? info : "");
	print_table_line(2, width_1_1, width_1_2);
	if (print_table_flag)
		printf("\n");

	print_table_line(2, width_2_1, width_2_2);
	if (print_table_flag)
		printf(format, width_2_1, "Limit", width_2_2, "Info");
	print_table_line(2, width_2_1, width_2_2);
	for (i = 0; i < nlimits_list; ++i) {
		if (print_table_flag)
			printf(format, width_2_1, limits_list[i].name, width_2_2, limits_list[i].info ? limits_list[i].info : "");
		else
			printf("%s %s\n", limits_list[i].name, limits_list[i].info ? limits_list[i].info : "");
	}
	print_table_line(2, width_2_1, width_2_2);
}

/*
 * Output accounting information for limit in rule.
 */
static void
print_limit_ac(void)
{
	char		*rule_info, *limit_info, *path, *format, *ptr;
	char		*byte_limit_str, *bytes_str, *byte_limit_conv_str, *bytes_conv_str;
	char		*buf = NULL;
	size_t		bufsize;
	u_quad_t	byte_limit, bytes;
	int		len, k;
	u_int		tmp, width_1, width_2, width_3, width_4;
	regex_t		reg_limit_line1, reg_limit_timestamp;
	FILE		*fp;
	struct stat	statbuf;
	struct timestamp {
		short		ts_is_set;
		struct tm	ts_tm;
		char		*date;
	}		ts_start, ts_zero, ts_reach, ts_expire, ts_exec1,
			ts_exec2, *ts_ptr;
	struct timestamp_arr {
		struct timestamp	*ptr;
		char			*status;
	}		ts_arr[6];
	char		*date_seq_msg = "date sequence in file %s %s";

	ts_arr[0].ptr = &ts_start;
	ts_arr[0].status = "Start";
	ts_arr[1].ptr = &ts_zero;
	ts_arr[1].status = "Zero";
	ts_arr[2].ptr = &ts_reach;
	ts_arr[2].status = "Reach";
	ts_arr[3].ptr = &ts_exec1;
	ts_arr[3].status = "Exec";
	ts_arr[4].ptr = &ts_expire;
	ts_arr[4].status = "Expire";
	ts_arr[5].ptr = &ts_exec2;
	ts_arr[5].status = "Exec";

	REGCOMP(limit_line1);
	REGCOMP(limit_timestamp);

	check_rule();

	if (acl_is_allowed(rule) == 0)
		warn_exit(no_rule_msg2, rule);

	if ( (ptr = get_info(rule, (char *)0)) != NULL) {
		if (asprintf(&rule_info, "%s", ptr) < 0)
			err_exit("asprintf");
	} else
		rule_info = NULL;
	limit_info = get_info(rule, limit);

	if (asprintf(&path, "%s/%s/" LIMITSDIR "/%s/" LIMITFILE, db_dir, rule, limit) < 0)
		err_exit("asprintf");

	if (stat(path, &statbuf) < 0) {
		if (errno != ENOENT)
			err_exit("stat(%s)", path);
		warn_exit("can't find limit \"%s\" in rule \"%s\".", limit, rule);
	}

	if (!S_ISREG(statbuf.st_mode))
		err_exit("%s should be regular file", path);
	if ( (fp = fopen(path, "r")) == NULL)
		err_exit("fopen(%s)", path);

	lock_db_file_until_end(path, fileno(fp), 0, SEEK_SET);

	len = readline(&buf, &bufsize, fp, path);
	if (len == 0)
		format_err("file %s", path);
	if (REGEXEC(limit_line1, buf) != 0)
		format_err("file %s (wrong first line)", path);

	if (sscanf(buf, "%qu %qu", &bytes, &byte_limit) != 2)
		err_exit("sscanf(%s, %%qu %%qu)", path);

	if (asprintf(&bytes_str, "%qu", bytes) < 0)
		err_exit("asprintf");
	if (asprintf(&byte_limit_str, "%qu", byte_limit) < 0)
		err_exit("asprintf");
	byte_limit_conv_str = convert_bytes(byte_limit);
	bytes_conv_str = convert_bytes(bytes);

	ts_start.ts_is_set = ts_zero.ts_is_set = ts_reach.ts_is_set =
	    ts_expire.ts_is_set = ts_exec1.ts_is_set = ts_exec2.ts_is_set = 0;

	for (k = 0; k < 7; ++k) {
		len = readline(&buf, &bufsize, fp, path);
		if (len == 0 && feof(fp) != 0) {
			if (k == 0)
				format_err("file %s (should have at least two lines)", path);
			break;
		}
		if (len > 0 && k == 6)
			format_err("file %s (file is too big)", path);
		if (len == 0) /* XXX readline() can't return 0 without EOF */
			err_exit("%s: read 0 bytes (%d line)", path, k + 2);
		if (REGEXEC(limit_timestamp, buf) != 0)
			format_err("file %s (%d line)", path, k + 2);

		ts_ptr = NULL;
		switch (*buf) {
		case LIMIT_STARTED:
			if (k != 0)
				break;
			if (!ts_start.ts_is_set)
				ts_ptr = &ts_start;
			break;
		case LIMIT_ZEROED:
			if (k != 1)
				break;
			if (!ts_zero.ts_is_set)
				ts_ptr = &ts_zero;
			break;
		case LIMIT_REACHED:
			if ((ts_zero.ts_is_set && k != 2) ||
			    (!ts_zero.ts_is_set && k != 1))
				break;
				if (!ts_reach.ts_is_set)
				ts_ptr = &ts_reach;
			break;
		case LIMIT_EXECUTED:
			if (k == 2 || k == 3) {
				if (!ts_exec1.ts_is_set)
					ts_ptr = &ts_exec1;
			} else if (k == 4 || k == 5)
				if (!ts_exec2.ts_is_set)
					ts_ptr = &ts_exec2;
			break;
		case LIMIT_EXPIRED:
			if ((!ts_zero.ts_is_set && !ts_exec1.ts_is_set && k != 2) ||
			    (!ts_zero.ts_is_set && ts_exec1.ts_is_set && k != 3) ||
			    (ts_zero.ts_is_set && !ts_exec1.ts_is_set && k != 3) ||
			    (ts_zero.ts_is_set && ts_exec1.ts_is_set && k != 4))
				break;
			if (!ts_expire.ts_is_set)
				ts_ptr = &ts_expire;
			break;
		default:
			format_err("file %s (line %d): unknown type of limit timestamp: '%c'", path, k + 2, *buf);
		}
		if (ts_ptr == NULL)
			format_err("file %s (line %d): wrong order of timestamps", path, k + 2);

		ts_ptr->ts_is_set = 1;

		if (sscanf(buf + 2, "%d.%d.%d/%d:%d:%d", &ts_ptr->ts_tm.tm_year,
		    &ts_ptr->ts_tm.tm_mon, &ts_ptr->ts_tm.tm_mday, &ts_ptr->ts_tm.tm_hour,
		    &ts_ptr->ts_tm.tm_min, &ts_ptr->ts_tm.tm_sec) != 6)
			err_exit("%s (line %d): sscanf(\"%s\", \"%%d.%%d.%%d/%%d:%%d:%%d\"): failed", path, k + 2, buf + 2);

		if (check_dates_flag && check_date(&ts_ptr->ts_tm) < 0)
			format_err(date_seq_msg, path, use_e_switch_msg);
	}

	if (fclose(fp) != 0) /* also release lock */
		err_exit("fclose(%s)", path);

	unlock_db();

	if (ts_zero.ts_is_set && ts_reach.ts_is_set)
		ts_zero.ts_is_set = 0;
	if (check_dates_flag) {
		ts_ptr = &ts_start;
		for (k = 0; k < 6; ++k) {
			if (ts_arr[k].ptr->ts_is_set) {
				if (tmcmp(&ts_arr[k].ptr->ts_tm, &ts_ptr->ts_tm) < 0)
					format_err(date_seq_msg, path, use_e_switch_msg);
				ts_ptr = ts_arr[k].ptr;
			}
		}
	}

	free(path);

	for (k = 0; k < 6; ++k) {
		if (ts_arr[k].ptr->ts_is_set)
			ts_arr[k].ptr->date = tm_str(&ts_arr[k].ptr->ts_tm);
	}

	width_1 = 4; /* Rule */
	width_2 = 9; /* Rule.Info */
	width_3 = 5; /* Limit */
	width_4 = 10; /* Limit.Info */

	if ( (tmp = strlen(rule)) > width_1)
		width_1 = tmp;
	if (rule_info)
		if ( (tmp = strlen(rule_info)) > width_2)
			width_2 = tmp;
	if ( (tmp = strlen(limit)) > width_3)
		width_3 = tmp;
	if (limit_info)
		if ( (tmp = strlen(limit_info)) > width_4)
			width_4 = tmp;

	print_table_line(4, width_1, width_2, width_3, width_4);
	format = "| %-*s | %-*s | %-*s | %-*s |\n";
	if (print_table_flag) {
		printf(format, width_1, "Rule", width_2, "Rule.Info", width_3, "Limit", width_4, "Limit.Info");
		print_table_line(4, width_1, width_2, width_3, width_4);
		printf(format, width_1, rule, width_2, rule_info ? rule_info : "", width_3, limit, width_4, limit_info ? limit_info : "");
		print_table_line(4, width_1, width_2, width_3, width_4);
	} else
		printf("%s %s\n%s %s\n", rule, rule_info ? rule_info : "", limit, limit_info ? limit_info : "");

	if (print_table_flag) {
		width_1 = 5; /* Bytes */
		if ( (tmp = strlen(bytes_str)) > width_1)
			width_1 = tmp;
		width_3 = strlen(bcnt_convert_msg[bcnt_convert_flag]); /* always > 'Cntr' */
		width_2 = width_4 = 10; /* Byte_limit, converted bcnt always < 'Byte_limit' */
		if ( (tmp = strlen(byte_limit_str)) > width_2)
			width_2 = tmp;
		if ( (tmp = strlen(bytes_conv_str)) > width_3)
			width_3 = tmp;
		if ( (tmp = strlen(byte_limit_conv_str)) > width_4)
			width_4 = tmp;
		printf("\n");
		print_table_line(4, width_1, width_2, width_3, width_4);
		printf(format, width_1, "Cntr", width_2, "Byte_limit", width_3, "Cntr", width_3, "Byte_limit");
		printf(format, width_1, "Bytes", width_2, "Bytes", width_3, bcnt_convert_msg[bcnt_convert_flag], width_4, bcnt_convert_msg[bcnt_convert_flag]);
		print_table_line(4, width_1, width_2, width_3, width_4);
		printf("| %*s | %*s | %*s | %*s |\n", width_1, bytes_str, width_2, byte_limit_str, width_3, bytes_conv_str, width_4, byte_limit_conv_str);
		print_table_line(4, width_1, width_2, width_3, width_4);
	} else
		printf("%s %s\n%s\n%s\n", bytes_str, byte_limit_str, bytes_conv_str, byte_limit_conv_str);

	width_1 = 6; /* Status */
	width_2 = 4; /* Date */
	for (k = 0; k < 6; ++k) {
		if (ts_arr[k].ptr->ts_is_set) {
			if ( (tmp = strlen(ts_arr[k].status)) > width_1)
				width_1 = tmp;
			if ( (tmp = strlen(ts_arr[k].ptr->date)) > width_2)
				width_2 = tmp;
		}
	}

	if (print_table_flag) {
		printf("\n");
		format = "| %-*s | %-*s |\n";
		print_table_line(2, width_1, width_2);
		printf(format, width_1, "Status", width_2, "Date");
		print_table_line(2, width_1, width_2);
	}
	for (k = 0; k < 6; ++k) {
		if (ts_arr[k].ptr->ts_is_set) {
			if (print_table_flag)
				printf(format, width_1, ts_arr[k].status, width_2, ts_arr[k].ptr->date);
			else
				printf("%s %s\n", ts_arr[k].status, ts_arr[k].ptr->date);
		}
	}
	print_table_line(2, width_1, width_2);
}

int
main(int argc, char *argv[])
{
	int		opt, show_all_flag = 0;

	opterr = 0;	/* don't allow getopt() prints own messages */
	while ( (opt = getopt(argc, argv, "hvatuend:l:r:i:I:AKMGT")) != -1) {
		switch (opt) {
		case 'v':
			show_version();
			return 0;
		case 'h':
			show_usage(argv[0]);
			return 0;
		case 'r':
			if (rule)
				err_exit("only one -r option can be used");
			rule = optarg;
			break;
		case 'l':
			if (limit)
				err_exit("only one -l option can be used");
			limit = optarg;
			break;
		case 'a':
			show_all_flag = 1;
			break;
		case 'd':
			db_dir = optarg;
			break;
		case 't':
			ac_per_day_flag = 0;
			break;
		case 'i':
			time_interval = optarg;
			break;
		case 'I':
			time_interval = optarg;
			use_second_timestamp_flag = 1;
			break;
		case 'u':
			set_ids_flag = 1;
			break;
		case 'e':
			check_dates_flag = 0;
			break;
		case 'A':
			bcnt_convert_flag = Full;
			break;
		case 'K':
			bcnt_convert_flag = Kbyte;
			break;
		case 'M':
			bcnt_convert_flag = Mbyte;
			break;
		case 'G':
			bcnt_convert_flag = Gbyte;
			break;
		case 'T':
			bcnt_convert_flag = Tbyte;
			break;
		case 'n':
			print_table_flag = 0;
			break;
		case '?':
			err_exit("invalid switch -%c", optopt);
		default:
			err_exit("getopt");
		}
	}

	if (optind < argc)
		err_exit("non-switch argument \"%s\"", argv[optind]);

	if (argc == 1)
		warn_exit("you should specify at least one switch.\nUse switch -h to view help message.");
	if (limit && !rule)
		err_exit("switch -l must be used with switch -r");
	if ((show_all_flag && rule && limit) || (limit && !ac_per_day_flag))
		err_exit("given combination of switches doesn't make sense.");

	get_my_ids();

	if (set_ids_flag || my_uid == 0) {
		/* -u or root invoked ipastat(8) */
		set_ids_flag = 1;
		if (setgid(my_gid) < 0)
			err_exit("setgid(%d)", my_gid);
		if (setuid(my_uid) < 0)
			err_exit("setuid(%d)", my_uid);
	}

	set_sighandlers();

	lock_db_read();

	if (show_all_flag) {
		/* -a */
		if (!rule)
			show_all_rules();
		else
			show_all_limits();
		return 0;
	}

	if (rule) {
		/* -r */
		if (limit) /* -l */
			print_limit_ac();
		else
			print_rule_ac();
	} else
		warn_exit("nothing to do with such options");
	return 0;
}

/*
 * Output version number (-v and -h).
 */
static void
show_version(void)
{
	printf(IPASTAT_MSG", version %s (%s)\n", Version, System);
}

/*
 * Output help message.
 */
static void
show_usage(char *envprogname)
{
	char	*ptr = strrchr(envprogname, '/');

	if (ptr != NULL)
		++ptr;
	else
		ptr = envprogname;
	show_version();
	printf("\
Usage: %s [-vhuatnAKMGT] [-d <db-directory>] [-r <rule>] [-l <limit>]\n\
	       [-i|I <time-interval>]\n\
   -d <db-direcory>\n\
   \tUse given <db-directory> instead of using default database directory\n\
   \t%s\n\
   -a\tPrint all names of rules which user can view or print all names of\n\
   \tlimits for the rule specified in the -r option.\n\
   -r <rule>\n\
   \tOutput accounting information for specified <rule>.\n\
   -l <limit>\n\
   \tOutput information for specified <limit>.\n\
   \tThis option can be used only with the -r option.\n\
   -A\tConvert values of byte counters to terabytes, gigabytes, megabytes,\n\
   \tkilobytes and bytes (convert as much as possible).\n\
   -K -M -G -T\n\
   \tConvert values of byte counters to kilobytes,\n\
   \tmegabytes (this is on by default), gigabytes or terabytes.\n\
   -t\tOutput accounting information for the rule per timestamp.\n\
   -i <time-interval>\n\
   \tSpecify <time-interval> for outputted accounting information, don't.\n\
   \tuse second timestamp.\n\
   -I <time-interval>\n\
   \tSpecify <time-interval> for outputted accounting information, but\n\
   \tuse second timestamp. (This is something different than the -i option,\n\
   \tplease read manual page ipastat(8) for more information.)\n\
   -u\tChange effective UID and GID to real UID and GID,\n\
   \tand don't use access control lists.\n\
   -e\tDon't check dates in accounting files, output all accounting\n\
   \tinformation even if there are errors in accounting files.\n\
   -n\tDon't indent information and print tables when output results.\n\
   \tUse this option if you want to parse output of this program.\n\
   -v\tShow version number and exit.\n\
   -h\tOutput this help message.\n", ptr, db_dir);
}

/*
 * Read line from fp, return number of characters in line or
 * -1 if error occured. *size is equal to #bytes previously allocated in *strp
 * by malloc(). If there is not enough space in *strp, then readline()
 * reallocates *strp and updates *size.
 *
 * This is wrapper for fgets() function with some extensions.
 */
int
readline(char **strp, size_t *size, FILE *fp, char *filename)
{
#define CHUNK_SIZE 100

#if (CHUNK_SIZE <= 1)
# error Macro CHUNK_SIZE should be greater than 1
#endif

	char		*s;
	size_t		can_nread;	/* #bytes to read in next fgets */
	size_t		len;		/* length of read string */
	size_t		nread = 0;	/* how many bytes we read */

	if (*strp == NULL) {
		if ( (*strp = malloc(CHUNK_SIZE)) == NULL)
			err_exit("malloc");
		*size = CHUNK_SIZE;
	}
	for (s = *strp, can_nread = *size; ;) {
		if (fgets(s, can_nread, fp) == NULL) {
			if (feof(fp) != 0)
				break;	/* EOF */
			err_exit("fgets(%s)", filename);
		}
		len = strlen(s);
		if (len > 0) {
			nread += len;
			if (s[len - 1] == '\n')
				break;
		}
		if ( (s = realloc(*strp, *size + CHUNK_SIZE)) == NULL)
			err_exit("realloc");
		*strp = s;
		*size += can_nread = CHUNK_SIZE;
		s += nread;
	}
	return nread;
}

/*
 * Get UID and GID of user, also get supplementary GIDs.
 */
void
get_my_ids(void)
{
	struct passwd	*passwdp;

	my_gid = getgid();
	if ( (my_uid = getuid()) == 0)
		return;
	if ( (passwdp = getpwuid(my_uid)) == NULL)
		return;

	ngid_list = 5;
	if ( (gid_list = malloc(ngid_list * sizeof *gid_list)) == NULL)
		err_exit("malloc");
	/* don't use getgroups(), because ipastat(8) is SUIDed/SGIDed */
	while (getgrouplist(passwdp->pw_name, my_gid, gid_list, &ngid_list) < 0) {
		ngid_list += 10;
		if ( (gid_list = realloc(gid_list, ngid_list * sizeof *gid_list)) == NULL)
			err_exit("realloc");
	}
}

/*
 * Convert full string to unsigned long integer.
 */
int
fullstrtoul(u_long *result, const char *nptr)
{
	char		*endptr;
	u_long		val;
	int		errno_save = errno;

	val = strtoul(nptr, &endptr, 10);
	if (errno != 0)
		return -1;
	errno = errno_save;
	if (*endptr != '\0')
		return -1;
	*result = val;
	return 0;
}

/*
 * Output message with error description and exit.
 */
void
err_exit(const char *message, ...)
{
	int		errno_save = errno;
	va_list		ap;

	va_start(ap, message);
	fflush(stdout);
	fprintf(stderr, "\nError: ");
	vfprintf(stderr, message, ap);
	if (errno_save)
		fprintf(stderr, ": %s", strerror(errno_save));
	fprintf(stderr, "\n\n");
	va_end(ap);
	exit(1);
}

/*
 * Output message and exit.
 */
void
warn_exit(const char *message, ...)
{
	va_list		ap;

	va_start(ap, message);
	printf("\nWarning: ");
	vprintf(message, ap);
	printf("\n\n");
	va_end(ap);
	exit(1);
}

/*
 * Output error message about wrong format of something and exit.
 */
void
format_err(const char *message, ...)
{
	va_list		ap;

	va_start(ap, message);
	fflush(stdout);
	fprintf(stderr, "\nError: wrong format of ");
	vfprintf(stderr, message, ap);
	fprintf(stderr, "\n\n");
	va_end(ap);
	exit(1);
}

/*
 * Convert given value of bytes as much as possible.
 */
char *
total_convert_bytes(u_quad_t a)
{
	char		*buf = NULL;
	char		*ptr;
	u_quad_t	t, g, m, k, b;

	t = a / TBYTE;
	a -= t * TBYTE;
	g = a / GBYTE;
	a -= g * GBYTE;
	m = a / MBYTE;
	a -= m * MBYTE;
	k = a / KBYTE;
	b = a - k * KBYTE;
	if (asprintf(&buf, "%quT %quG %quM %quK %quB", t, g, m, k , b) < 0)
		err_exit("asprintf");
	ptr = buf;
	if (t == 0) {
		ptr = skip_first_space(ptr);
		if (g == 0) {
			ptr = skip_first_space(ptr);
			if (m == 0) {
				ptr = skip_first_space(ptr);
				if (k == 0)
					ptr = skip_first_space(ptr);
			}
		}
	}
	if (b == 0) {
		last_space_to_nul(ptr);
		if (k == 0) {
			last_space_to_nul(ptr);
			if (m == 0) {
				last_space_to_nul(ptr);
				if (g == 0)
					last_space_to_nul(ptr);
			}
		}
	}
	return *ptr == '\0' ? "0B" : ptr;
}

void
sig_alrm(int signo)
{
	longjmp(env_alrm, 1); 
}

void
set_sighandlers(void)
{
	struct sigaction	act;

	if (signal(SIGTSTP, SIG_IGN) == SIG_ERR)
		err_exit("signal(SIGTSTP)");
	if (signal(SIGTTOU, SIG_IGN) == SIG_ERR)
		err_exit("signal(SIGTTOU)");
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	act.sa_handler = sig_alrm;
	if (sigaction(SIGALRM, &act, (struct sigaction *)0) < 0)
		err_exit("sigaction(SIGALRM)");
}

void
lock_db_read(void)
{
	if (asprintf(&lockdbpath, "%s/%s", db_dir, lockdbfile) < 0)
		err_exit("asprintf");

	if ( (db_fd = open(lockdbpath, O_RDONLY)) < 0)
		err_exit("open(%s)", lockdbpath);

	if (setjmp(env_alrm) != 0)
		warn_exit("can't lock file %s (database) during %d seconds", lockdbpath, LOCK_WAIT_TIME);

	alarm(LOCK_WAIT_TIME);
	if (readw_lock(db_fd, 0, SEEK_SET, 0) < 0)
		err_exit("fcntl(F_SETLKW)");
	alarm(0);
}

/*
 * Unlock whole database.
 */
void
unlock_db(void)
{
	if (close(db_fd) < 0)
		err_exit("close(%s)", lockdbfile);
}

/*
 * Obtain "read" lock on file from start to the end of file. Starting offset
 * is measured from the position specified by whence parameter (SEEK_CUR,
 * SEEK_SET and SEEK_END).
 */
void
lock_db_file_until_end(char *file, int fd, int start, short whence)
{
	if (setjmp(env_alrm) != 0)
		warn_exit("can't lock file %s during %d seconds", file, LOCK_WAIT_TIME);
	alarm(LOCK_WAIT_TIME);
	if (readw_lock(fd, start, whence, 0) < 0)
		err_exit("fcntl(%s, F_SETLKW)", file);
	alarm(0);
}
